/*******************************************************************************/
/* Elevator model                                                              */
/* Author: Peter Schrammel                                                     */
/*******************************************************************************/
/* modelled after                                                              */
/*     Meinke, K., Sindhu, M.A.: Incremental learning-based testing for 
       reactive systems. In: TAP. Volume 6706 of LNCS. (2011) 134–151          */
/* if you reuse this code, please cite also                                    */
/*      Peter Schrammel, Tom Melham, Daniel Kroening. Chaining Test Cases for 
        Reactive System Testing. In Testing Software and Systems
        LNCS Volume 8254, 2013, pp 133-148                                     */
/*******************************************************************************/

#include <assert.h>

typedef struct state {
  int stop;
  int cl;
  int w1;
  int w2;
  int w3;
  int at;
  int up;
  int sw2; //stop at 2
} t_state;

typedef struct io {
  int tick; 
  int c1; 
  int c2; 
  int c3; 
} t_io;

void init(t_state *s) {
  s->stop = 1;
  s->cl = 1;
  s->w1 = 0;
  s->w2 = 0;
  s->w3 = 0;
  s->at = 1;
  s->up = 1;
  s->sw2 = 0;
}

void compute(t_io* i, t_state *s) {
  if(s->stop) {
    if(!s->w1 && s->at==1) {
      if(!s->w2 && !s->w3) {
        if(s->cl && i->c1) { s->cl = 0; return;}
        if(i->c2) { s->w2 = 1; s->stop = 0; s->cl = 1; s->up = 1; s->at = 0; s->sw2 = 1; return; }
        if(i->c3) { s->w3 = 1; s->stop = 0; s->cl = 1; s->up = 1; s->at = 0; s->sw2 = 0; return; }
        if(!s->cl && i->tick) { s->cl = 1; return; }
      }
      if(!s->w2 && s->w3 && !s->cl) {
        if(i->c3 || i->tick) { s->stop = 0; s->cl = 1; s->up = 1; s->at = 0; s->sw2 = 0; return; }
        if(i->c2) { s->w2 = 1; s->stop = 0; s->cl = 1; s->up = 1; s->at = 0; s->sw2 = 1; return; }
      }
      if(s->w2 && !s->w3 && !s->cl) {
        if(i->c2 || i->tick) { s->stop = 0; s->cl = 1; s->up = 1; s->at = 0; s->sw2 = 1; return; }
        if(i->c3) { s->w3 = 1; s->stop = 0; s->cl = 1; s->up = 1; s->at = 0; s->sw2 = 0; return; }
      }
      if(s->w2 && s->w3 && !s->cl) {
        if(i->c2 || i->c3 || i->tick) { s->cl = 1; s->stop = 0; s->up = 1; s->at = 0; s->sw2 = 1; return; }
      }
    }
    else if(!s->w2 && s->at==2) {
      if(!s->w1 && !s->w3) {
        if(s->cl && i->c2) { s->cl = 0; return; }
        if(i->c1) { s->w1 = 1; s->cl = 1; s->stop = 0; s->up = 0; s->at = 0; s->sw2 = 1; return; }
        if(i->c3) { s->w3 = 1; s->cl = 1; s->stop = 0; s->up = 1; s->at = 0; s->sw2 = 1; return; }
        if(!s->cl && i->tick) { s->cl = 1; return; }
      }
      if(s->up && !s->w1 && s->w3 && !s->cl) {
        if(i->c1) { s->w1 = 1; s->stop = 0; s->cl = 1; s->at = 0; s->sw2 = 1; return; }
        if(i->c3 || i->tick) { s->stop = 0; s->cl = 1; s->at = 0; s->sw2 = 1; return; }
      }
      if(!s->up && !s->w1 && s->w3 && !s->cl) {
        if(i->c3 || i->tick) { s->stop = 0; s->cl = 1; s->up = 1; s->at = 0; s->sw2 = 1; return; }
        if(i->c1) { s->w1 = 1; s->stop = 0; s->cl = 1; s->at = 0; s->sw2 = 1; return; }
      }
      if(s->up && s->w1 && !s->w3 && !s->cl) {
        if(i->c1 || i->tick) { s->stop = 0; s->cl = 1; s->up = 0; s->at = 0; s->sw2 = 1; return; }
        if(i->c3) { s->w3 = 1; s->stop = 0; s->cl = 1; s->at = 0; s->sw2 = 1; return; }
      }
      if(!s->up && s->w1 && !s->w3 && !s->cl) {
        if(i->c1 || i->tick) { s->stop = 0; s->cl = 1; s->at = 0; s->sw2 = 1; return; }
        if(i->c3) { s->w3 = 1; s->stop = 0; s->cl = 1; s->at = 0; s->sw2 = 1; return; }
      }
      if(s->up && s->w1 && s->w3) {
        if(i->c1 || i->c3 || i->tick) { s->stop = 0; s->cl = 1; s->at = 0; s->sw2 = 1; return; }
      }
      if(!s->up && s->w1 && s->w3) {
        if(i->c1 || i->c3 || i->tick) { s->stop = 0; s->cl = 1; s->at = 0; s->sw2 = 1; return; }
      }
    }
    else if(!s->w3 && s->at==3) {
      if(!s->w1 && !s->w2) {
        if(s->cl && i->c3) { s->cl = 0; return; }
        if(i->c1) { s->w1 = 1; s->cl = 1; s->stop = 0; s->up = 0; s->at = 0; s->sw2 = 0; return; }
        if(i->c2) { s->w2 = 1; s->cl = 1; s->stop = 0; s->up = 0; s->at = 0; s->sw2 = 1; return; }
	if(!s->cl && i->tick) { s->cl = 1; return; }
      }
      if(!s->w1 && s->w2 && !s->cl) {
        if(i->c1) { s->w1 = 1; s->stop = 0; s->up = 0; s->at = 0; s->sw2 = 0; return; }
        if(i->c2 || i->tick) { s->stop = 0; s->up = 0; s->at = 0; s->sw2 = 1; return; }
      }
      if(s->w1 && !s->w2 && !s->cl) {
        if(i->c1 || i->tick) { s->stop = 0; s->up = 0; s->cl = 1; s->at = 0; s->sw2 = 0; return; }
        if(i->c2) { s->w2 = 1; s->stop = 0; s->up = 0; s->cl = 1; s->at = 0; s->sw2 = 1; return; }
      }
      if(s->w1 && s->w2) {
        if(i->c1 || i->c2 || i->tick) { s->cl = 1; s->stop = 0; s->up = 0; s->sw2 = 1; return; }
      }
    }
  }
  else if(s->cl && s->at!=1 && s->at!=2 && s->at!=3) { //go
    if(s->up) { //up
      if(!s->sw2) {
        if(!s->w2 && s->w3) {
          if(!s->w1) {
            if(i->c1) { s->w1 = 1; return; }
            if(i->c2) { s->w2 = 1; s->sw2 = 1; return; }
            if(i->c3 || i->tick) { s->sw2 = 1; return; }
  	  }
          if(s->w1) {
            if(i->c1 || i->c3 || i->tick) { s->sw2 = 1; return; }
            if(i->c2) { s->w2 = 1; s->sw2 = 1; return; }
  	  }
        }
      }
      else {
	if(s->w2) {
	  if(s->w1 && !s->w3) {
            if(i->c3) { s->w3 = 1; return; }
            if(i->c2 || i->tick) { s->w1 = 1; s->w2 = 0; s->at = 2; s->stop = 1; s->cl = 0; return; }
	  }
	  if(s->w1 && s->w3) {
            if(i->c1 || i->c2 || i->c3 || i->tick) { s->w2 = 0; s->at = 2; s->stop = 1; s->cl = 0; return; }
	  }
	  if(!s->w1 && !s->w3) {
            if(i->c1) { s->w1 = 1; return; }
            if(i->c2 || i->tick) {  s->w2 = 0; s->at = 2; s->stop = 1; s->cl = 0; return; }
	  }
	  if(!s->w1 && s->w3) {
            if(i->c1) { s->w1 = 1; return; }
            if(i->c2 || i->c3 || i->tick) { s->w2 = 0; s->at = 2; s->stop = 1; s->cl = 0; return; }
	  }
	}
	if(s->w3) {
	  if(s->w1 && !s->w2) {
            if(i->c2) { s->w2 = 1; return; }
            if(i->c1 || i->c3 || i->tick) { s->w3 = 0; s->at = 3; s->stop = 1; s->cl = 0; return; }
	  }
	  if(s->w1 && s->w2) {
            if(i->c1 || i->c2 || i->c3 || i->tick) { s->w3 = 0; s->at = 3; s->stop = 1; return; }
	  }
	  if(!s->w1 && !s->w2) {
            if(i->c1) { s->w1 = 1; return; }
            if(i->c2) { s->w2 = 1; return; }
            if(i->c3 || i->tick) { s->w3 = 0; s->at = 3; s->stop = 1; s->cl = 0; return; }
	  }
	  if(!s->w1 && s->w2) {
            if(i->c1) { s->w1 = 1; return; }
            if(i->c2 || i->c3 || i->tick) { s->w3 = 0; s->at = 3; s->stop = 1; s->cl = 0; return; }
	  }      
	}
      }
    }
    else { //down
      if(!s->sw2) {
	if(s->w1 && !s->w2) {
	  if(!s->w3) {
            if(i->c2) { s->w2 = 1; s->sw2 = 1; return; }
            if(i->c3) { s->w3 = 1; return; }
            if(i->c1 || i->tick) { s->sw2 = 1; return; }
	  }
	  if(s->w3) {
            if(i->c2) { s->w2 = 1; s->sw2 = 1; return; }
            if(i->c1 || i->c3 || i->tick) { s->sw2 = 1; return; }
	  }
	}
      }
      else {
	if(s->w2) {
	  if(s->w1 && !s->w3) {
            if(i->c3) { s->w3 = 1; return; }
            if(i->c1 || i->c2 || i->tick) { s->w2 = 0; s->at = 2; s->stop = 1; s->cl = 0; return; }
	  }
	  if(s->w1 && s->w3) {
            if(i->c1 || i->c2 || i->c3 || i->tick) { s->w2 = 0; s->at = 2; s->stop = 1; s->cl = 0; return; }
	  }
	  if(!s->w1 && !s->w3) {
            if(i->c1) { s->w1 = 1; return; }
            if(i->c3) { s->w3 = 1; return; }
            if(i->c2 || i->tick) { s->w2 = 0; s->at = 2; s->stop = 1; s->cl = 0; return; }
	  }
	  if(!s->w1 && s->w3) {
            if(i->c1) { s->w1 = 1; return; }
            if(i->c2 || i->c3 || i->tick) { s->w2 = 0; s->at = 2; s->stop = 1; s->cl = 0; return; }
	  }
	}
	if(s->w1) {
	  if(s->w2 && !s->w3) {
            if(i->c3) { s->w3 = 1; return; }
            if(i->c1 || i->c2 || i->tick) { s->w1 = 0; s->at = 1; s->stop = 1; s->cl = 0; return; } 
	  }
	  if(s->w2 && s->w3) {
            if(i->c1 || i->c2 || i->c3 || i->tick) { s->w1 = 0; s->at = 1; s->stop = 1; s->cl = 0; return; }
	  }
	  if(!s->w2 && !s->w3) {
            if(i->c2) { s->w2 = 1; return; }
            if(i->c3) { s->w3 = 1; return; }
            if(i->c1 || i->tick) { s->w1 = 0; s->at = 1; s->stop = 1; s->cl = 0; return; }
	  }
	  if(!s->w2 && s->w3) {
            if(i->c2) { s->w2 = 1; return; }
            if(i->c1 || i->c3 || i->tick) { s->w1 = 0; s->at = 1; s->stop = 1; s->cl = 0; return; }
	  }      
	}
      }
    }
  }
}

#ifdef CBMC

void input_assume(t_io* i, t_state* s) {
  __CPROVER_assume((0<=i->tick) && (i->tick<=1) &&
                   (0<=i->c1) && (i->c1<=1) &&
                   (0<=i->c2) && (i->c2<=1) &&
                   (0<=i->c3) && (i->c3<=1) &&
                   (
                     (i->c1 || i->c2 || i->c3) && !i->tick ||
                     !i->c1 && !i->c2 && !i->c3 && i->tick
		    )
                  );
}


extern t_io nondet_input();

int main(int argc, char** argv) 
{
  t_state s;
  init(&s);
  
  while(1)
  {
    //step
    t_io i = nondet_input();
    input_assume(&i,&s);
    t_state s_old = s;
    compute(&i,&s);

    //properties

    //stopped, no request => doors closed
    if(s_old.stop && s_old.cl && !s_old.w1 && !s_old.w2 && 
       !s_old.w3 && !i.c1 && !i.c2 && !i.c3)
      __CPROVER_assert(s.cl,"c::prop1");

    //stopped, doors open, any request => doors close
    if(s_old.stop && s_old.cl && !s_old.w1 && !s_old.w2 && !s_old.w3 && 
       (s_old.at==1 && !i.c1 && (i.c2 || i.c3) ||
	s_old.at==2 && !i.c2 && (i.c1 || i.c3) ||
	s_old.at==3 && !i.c3 && (i.c1 || i.c2)))
      __CPROVER_assert(!s.stop && s.up,"c::prop2");

    //stopped, doors closed, request at respective floor => doors open
    if(s_old.stop && s_old.cl && !s_old.w1 && !s_old.w2 && !s_old.w3 && 
       (s_old.at==1 && i.c1 || s_old.at==2 && i.c2 || s_old.at==3 && i.c3))
      __CPROVER_assert(!s.cl,"c::prop3");

    //arriving at requested floor => stop and doors open
    if(!s_old.stop && i.tick && 
     (s_old.w1 && !s_old.w2 && !s_old.w3 && s_old.sw2 && !s_old.up  ||
      !s_old.w1 && s_old.w2 && !s_old.w3 && s_old.sw2  ||
      !s_old.w1 && !s_old.w2 && s_old.w3 && s_old.sw2 && s_old.up))
      __CPROVER_assert(s.stop && !s.cl,"c::prop4");

    //not stopped, request => request stored
    if(!s_old.stop && s_old.cl && !s_old.w1 && i.c1)
      __CPROVER_assert(s.w1,"c::prop5");

    //not stopped, request => request stored
    if(!s_old.stop && s_old.cl && !s_old.w2 && i.c2)
      __CPROVER_assert(s.w2,"c::prop6");

    //not stopped, request => request stored
    if(!s_old.stop && s_old.cl && !s_old.w3 && i.c3)
      __CPROVER_assert(s.w3,"c::prop7");

    //doors open, request at respective floor => doors open
    if(s_old.stop && s_old.cl && s_old.at==1 && i.c1)
      __CPROVER_assert(s.stop && !s.cl && !s.w1,"c::prop8");

    //doors open, request at respective floor => doors open
    if(s_old.stop && s_old.cl && s_old.at==2 && i.c2)
      __CPROVER_assert(s.stop && !s.cl && !s.w2,"c::prop9");

    //doors open, request at respective floor => doors open
    if(s_old.stop && s_old.cl && s_old.at==3 && i.c3)
      __CPROVER_assert(s.stop && !s.cl && !s.w3,"c::prop10");

    //stopped, doors open, any request => doors close
    if(s_old.stop && s_old.at==1 && !s_old.cl && !s_old.w1 && 
       !s_old.w2 && 
       !s_old.w3 && !i.c1 && (i.c2 || !i.c3))
      __CPROVER_assert(s.cl,"c::prop11");

    //stopped, doors open, any request => doors close
    if(s_old.stop && s_old.at==2 && !s_old.cl && !s_old.w1 && 
       !s_old.w2 && 
       !s_old.w3 && (i.c1 || i.c3) && !i.c2)
      __CPROVER_assert(s.cl,"c::prop12");

    //stopped, doors open, any request => doors close
    if(s_old.stop && s_old.at==3 && !s_old.cl && !s_old.w1 && 
       !s_old.w2 && 
       !s_old.w3 && (i.c1 || i.c2) && !i.c3)
      __CPROVER_assert(s.cl,"c::prop13");

    //stopped, doors closed, request at respective floor => doors open
    if(s_old.stop && s_old.at==1 && s_old.cl && !s_old.w1 && 
       !s_old.w2 && !s_old.w3 && i.c1)
      __CPROVER_assert(!s.cl,"c::prop14");

    //stopped, doors closed, request at respective floor => doors open
    if(s_old.stop && s_old.at==2 && s_old.cl && !s_old.w1 && 
       !s_old.w2 && !s_old.w3 && i.c2)
      __CPROVER_assert(!s.cl,"c::prop15");

    //stopped, doors closed, request at respective floor => doors open
    if(s_old.stop && s_old.at==3 && s_old.cl && !s_old.w1 && 
       !s_old.w2 && !s_old.w3 && i.c3)
      __CPROVER_assert(!s.cl,"c::prop16");

    //arriving at requested floor => stop and doors open
    if(!s_old.stop && s_old.cl && !s_old.up &&
     s_old.w1 && !s_old.w2 && !s_old.w3 && s_old.sw2 && !i.c2 && !i.c3)
      __CPROVER_assert(s.stop && !s.cl && s.at==1,"c::prop17");

    //arriving at requested floor => stop and doors open
    if(!s_old.stop && s_old.cl && !s_old.up &&
     !s_old.w1 && s_old.w2 && !s_old.w3 && s_old.sw2 && !i.c1 && !i.c3)
      __CPROVER_assert(s.stop && !s.cl && s.at==2,"c::prop18");

    //arriving at requested floor => stop and doors open
    if(!s_old.stop && s_old.cl && s_old.up &&
     !s_old.w1 && s_old.w2 && !s_old.w3 && s_old.sw2 && !i.c1 && !i.c3)
      __CPROVER_assert(s.stop && !s.cl && s.at==2,"c::prop19");

    //arriving at requested floor => stop and doors open
    if(!s_old.stop && s_old.cl && s_old.up &&
     !s_old.w1 && !s_old.w2 && s_old.w3 && s_old.sw2 && !i.c1 && !i.c2)
      __CPROVER_assert(s.stop && !s.cl && s.at==3,"c::prop20");

    //up, request => request stored
    if(!s_old.stop && s_old.up && s_old.cl && !s_old.w1 && i.c1)
      __CPROVER_assert(s.w1,"c::prop21");

    //up, request => request stored
    if(!s_old.stop && s_old.up && s_old.cl && !s_old.w2 && i.c2)
      __CPROVER_assert(s.w2,"c::prop22");

    //up, request => request stored
    if(!s_old.stop && s_old.up && s_old.cl && !s_old.w3 && i.c3)
      __CPROVER_assert(s.w3,"c::prop23");

    //down, request => request stored
    if(!s_old.stop && !s_old.up && s_old.cl && !s_old.w1 && i.c1)
      __CPROVER_assert(s.w1,"c::prop24");

    //down, request => request stored
    if(!s_old.stop && !s_old.up && s_old.cl && !s_old.w2 && i.c2)
      __CPROVER_assert(s.w2,"c::prop25");

    //down, request => request stored
    if(!s_old.stop && !s_old.up && s_old.cl && !s_old.w3 && i.c3)
      __CPROVER_assert(s.w3,"c::prop26");

    //final
    if(s.stop && (s.at==1) && s.cl && !s.w1 && 
       !s.w2 && !s.w3) 
      __CPROVER_assert(1,"c::final");

   }

  return 0;
}
#endif


