/*******************************************************************\

Module: Bounded Model Checking for ANSI-C + HDL

Author: Daniel Kroening, kroening@kroening.com

\*******************************************************************/

#ifndef CHAINCOVER_BMC_H
#define CHAINCOVER_BMC_H

#include <list>
#include <map>
#include <exception>

#include <util/hash_cont.h>
#include <util/options.h>

#include <solvers/prop/prop.h>
#include <solvers/prop/prop_conv.h>
#include <solvers/sat/cnf.h>
#include <solvers/sat/satcheck.h>
#include <solvers/smt1/smt1_dec.h>
#include <solvers/smt2/smt2_dec.h>
#include <langapi/language_ui.h>
#include <goto-symex/symex_target_equation.h>

#include "symex_bmc.h"
#include "bv_cbmc.h"


class bmct:public messaget
{
public:
  bmct(
    const optionst &_options,
    const symbol_tablet &_symbol_table,
    message_handlert &_message_handler,
    prop_convt& _prop_conv):
    messaget(_message_handler),
    generate_error_trace(true),
    equation(ns),
    symex(ns, new_symbol_table, equation,_prop_conv),
    ns(_symbol_table, new_symbol_table),
    options(_options),
    ui(ui_message_handlert::PLAIN)
  {
    symex.constant_propagation=options.get_bool_option("propagation");
  }
 
  virtual bool run_symex(const goto_functionst &goto_functions, 
                         const goto_programt &goto_program);
  virtual bool run(const goto_functionst &goto_functions, 
                   const goto_programt &goto_program);
  virtual ~bmct() { }

  // additional stuff   
  expr_listt bmc_constraints;  
  
  void set_ui(language_uit::uit _ui) { 
    ui=_ui; 
    symex.set_ui(ui);
  }

  bool generate_error_trace;
  goto_tracet goto_trace;
  symex_target_equationt equation;
  symex_bmct symex;

  virtual void setup_unwind();

  virtual void do_unwind_module(
    decision_proceduret &decision_procedure);

  namespacet ns;

  virtual void error_trace(
    const prop_convt &prop_conv);

 
protected:
  const optionst &options;  
  symbol_tablet new_symbol_table;
 
  // use gui format
  language_uit::uit ui;
  
  virtual decision_proceduret::resultt
    run_decision_procedure(prop_convt &prop_conv);
    
  virtual bool decide(prop_convt &prop_conv);
    
  void do_conversion(prop_convt &solver);
  
  virtual void show_program();
  virtual void report_success();
  virtual void report_failure();


};

void build_goto_trace_with_all_assertions(
  const symex_target_equationt &target,
  const prop_convt &prop_conv,
  const namespacet &ns,
  goto_tracet &goto_trace);


#endif
