/*******************************************************************\

Module: Command Line Parsing

Author: Daniel Kroening, kroening@kroening.com

\*******************************************************************/

#ifndef CHAINCOVER_PARSEOPTIONS_H
#define CHAINCOVER_PARSEOPTIONS_H

#include <util/ui_message.h>
#include <util/parseoptions.h>

#include <langapi/language_ui.h>

#include "xml_interface.h"

class bmct;
class goto_functionst;
class optionst;

#define CBMC_OPTIONS \
  "(program-only)(function):(preprocess)(slice-by-trace):" \
  "(no-simplify)(unwind):(unwindset):(slice-formula)" \
  "(unwind-max):" \
  "(debug-level):(no-propagation)(no-simplify-if)" \
  "(bounds-check)(outfile):(pointer-check)" \
  "(document-subgoals)D:I:(depth):" \
  "(div-by-zero-check)(no-unwinding-assertions)" \
  "(signed-overflow-check)(unsigned-overflow-check)" \
  "(partial-loops)" \
  "(xml-ui)(xml-interface)(vcd):" \
  "(cvc)(smt1)(smt2)(boolector)(yices)(z3)(opensmt)(mathsat)(fpa)" \
  "(no-pretty-names)(beautify)" \
  "(floatbv)(fixedbv)(no-assertions)(no-assumptions)(nan-check)" \
  "(dimacs)(refine)" \
  "(16)(32)(64)(LP64)(ILP64)(LLP64)(ILP32)(LP32)" \
  "(little-endian)(big-endian)" \
  "(show-goto-functions)(show-value-sets)(show-loops)" \
  "(show-symbol-table)(show-vcc)(show-properties)(claim):" \
  "(error-label):(verbosity):(no-library)" \
  "(version)" \
  "(cover-assertions)" \
  "(memory-model):" \
  "(i386-linux)(i386-macos)(i386-win32)(win32)(winx64)(gcc)" \
  "(ppc-macos)(unsigned-char)" \
  "(arrays-uf-always)(arrays-uf-never)" \
  "(string-abstraction)(no-arch)" \
  "(round-to-nearest)(round-to-plus-inf)(round-to-minus-inf)(round-to-zero)" \
  "(chain-properties)(clingo)(path-constraints)(no-multi-chain)" \
  "(config):(output-config)" \
  "(branch)" \
  "(chain)(no-chain)(incremental-check):(output-tests):" \
  "(do-not-minimize)" \
  "(decide)" // legacy, and will eventually disappear 
//(output-harness):

class chaincover_parseoptionst:
  public parseoptions_baset,
  public xml_interfacet,
  public language_uit
{
public:
  virtual int doit();
  virtual void help();

  chaincover_parseoptionst(int argc, const char **argv);
  chaincover_parseoptionst(
    int argc,
    const char **argv,
    const std::string &extra_options);

protected:
  virtual void register_languages();

  virtual void get_command_line_options(optionst &options);

  virtual bool doit(const optionst &options,
                    goto_functionst &goto_functions);

  virtual bool get_goto_program(
    const optionst &options,
    goto_functionst &goto_functions);

  virtual bool process_goto_program(
    const optionst &options,
    goto_functionst &goto_functions);
    
  bool set_properties(goto_functionst &goto_functions);
  
  void eval_verbosity();
  
 
  void preprocessing();
};

#endif
