#ifndef CHAINCOVER_CONFIG_PARSER_XML_H
#define CHAINCOVER_CONFIG_PARSER_XML_H

#include <fstream>
#include <iostream>
#include <util/message.h>
#include <goto-programs/goto_functions.h>
#include <util/std_expr.h>
#include <util/string2int.h>
#include <xmllang/xml_parser.h>

#include "config_parser_base.h"

#define DEFAULT_CHAIN_BOUND 0 //no bound, search optimum
#define DEFAULT_FINAL "c::final"
#define DEFAULT_MAIN "c::main"

class config_parser_xmlt : public config_parser_baset {
 public:
  config_parser_xmlt(std::string _filename) : 
    config_parser_baset(_filename) {}

  virtual void parse() 
  {
    null_message_handlert message_handler;
    xmlt xml;
    parse_xml(filename, message_handler, xml);
    for(xmlt::elementst::const_iterator it=xml.elements.begin();
        it!=xml.elements.end(); it++) {
      if(it->name=="chain") {
        for(xmlt::elementst::const_iterator itch=it->elements.begin();
            itch!=it->elements.end(); itch++) {
          if(itch->name=="main") main = itch->data;
          if(itch->name=="final") final = itch->data;
          if(itch->name=="chain-bound") chain_bound = 
            safe_string2unsigned(itch->data.c_str());
          if(itch->name=="testgoals") {
            for(xmlt::elementst::const_iterator itpr=itch->elements.begin();
                itpr!=itch->elements.end(); itpr++) {
              if(itpr->name=="testgoal") {
                properties.insert(itpr->data);
  	      }
            }
          }
	}
      }
    }
  }

  virtual void infer(const goto_functionst& goto_functions, 
		     const namespacet &ns) 
  {
    propertiest properties; 
    unsigned chain_bound = DEFAULT_CHAIN_BOUND;
    propertyt final = DEFAULT_FINAL; 

    forall_goto_functions(it, goto_functions)
    {
      std::string fname = id2string(it->first);	
      if(fname.find("__CPROVER")!=std::string::npos) continue;
      //use all assertions as properties
      forall_goto_program_instructions(i, it->second.body)
      {
        if(i->is_assert())
	{
	  std::string comment = id2string(i->location.get_comment());
          if(comment.find("final")!=std::string::npos) final = comment;
          else if(comment!="") properties.insert(comment);
	}
      }
    }

    //write to file
    std::ofstream outfile(filename.c_str());
    if (!outfile.is_open()) throw ("ERROR: cannot open file "+filename);
    outfile << "<chaincover-config>" << std::endl;
    outfile << "  <chain>" << std::endl;
    outfile << "    <main>" << DEFAULT_MAIN << "</main>" << std::endl;
    outfile << "    <final>" << final << "</final>" << std::endl;
    outfile << "    <chain-bound>" << chain_bound 
      << "</chain-bound>" << std::endl;
    outfile << "    <testgoals>" << std::endl;
    for(propertiest::iterator p = properties.begin(); 
      p!=properties.end(); p++) {
      outfile << "      <testgoal>" << *p << "</testgoal>" << std::endl;
    }
    outfile << "    </testgoals>" << std::endl;
    outfile << "  </chain>" << std::endl;
    outfile << "</chaincover-config>" << std::endl;
    outfile.close();
  }

  propertyt main;

};

#endif
