#include "reach_bmc.h"
#include <iostream>

#define INIT_VAR "__init__"

void translate_func(irep_idt fname, std::map<exprt,exprt>& sequ, std::map<exprt,exprt>& lequ, bmccontextt* context) {
  const goto_functionst::goto_functiont *f;
  goto_functionst::function_mapt::const_iterator f_it = context->goto_functions->function_map.find(fname);
  if(f_it==context->goto_functions->function_map.end()) throw "ERROR: function not found";
  f = &(f_it->second);

  //target conditions
  //  std::map<irep_idt,exprt> cond;

  // <expr> = 
  //  local declaration: add to locals
  //  associate condition c to each statement
  //  IF e THEN GOTO x; y: c, x(c/\e), y(c/\-e)
  //  assignment z = e for variable z: if c then e else
  //  GOTO x: x(c)
  //  e ? TRUE : FALSE: e
  //  TRUE: true, FALSE: false, !: not, ==: =, _Bool: bool
  //  if bool var: 0: false, 1: true
  //  dead: ignore
  //  function call f: f(c)

  for(goto_programt::instructionst::const_iterator 
      it=(f->body).instructions.begin(); 
       it!=(f->body).instructions.end(); it++) {
    switch(it->type) {
      /*     case DECL: 
        code_declt i = to_code_decl(it->code);
	lequ.insert(i.symbol(),i.initialiser()); 
        break;*/
      case ASSIGN: {
        code_assignt i = to_code_assign(it->code);
        std::map<exprt,exprt>::iterator lit = lequ.find(i.lhs());
        if(lit == lequ.end()) sequ.insert(std::pair<exprt, exprt>(i.lhs(),i.rhs()));
        else lit->second = i.rhs();
        break; }
	/*      case FUNCTION_CALL: 
          translate_func(it->,sequ,lequ);
	  break;*/
      default: break;
	//	throw "ABORT: unexpected statement";
    }
  }
}

//std::map<irep_idt,string> typemap
void reach_bmct::output_goto_nbac(irep_idt init, irep_idt step, irep_idt input_assume, std::ostream& out) {

  //argument names //todo: check
  //  irep_idt iname = fstep->type.arguments()[0].get_identifier();
  //  irep_idt sname = fstep->type.arguments()[1].get_identifier();

  //input variables
  std::set<exprt> invar;

  symbol_exprt initvar;
  initvar.set_identifier(INIT_VAR);
  initvar.type()=bool_typet();

  //state equations
  std::map<exprt,exprt> sequ_init;
  std::map<exprt,exprt> sequ_step;
  //local definitions
  std::map<exprt,exprt> lequ;

  //build initexpr and stepexpr
  translate_func(init,sequ_init,lequ,context);
  translate_func(step,sequ_step,lequ,context);
  
  //ite(initvar,initexpr,stepexpr)
  std::map<exprt,exprt> sequ;
  for(std::map<exprt,exprt>::iterator it=sequ_init.begin(); it!=sequ_init.end(); it++) {
    if_exprt e;
    e.cond() = initvar;
    e.true_case() = it->second;
    std::map<exprt,exprt>::iterator sit = sequ_step.find(it->first);
    if(sit != sequ_step.end()) e.false_case() = sit->second;
    else e.false_case() = it->first;
    sequ.insert(std::pair<exprt, exprt>(it->first,e));
    }
  sequ.insert(std::pair<exprt, exprt>(initvar,false_exprt()));

  //print
  out << "state" << std::endl;
  for(std::map<exprt,exprt>::iterator it=sequ.begin(); it!=sequ.end(); it++) {
    out << it->first << " : " /* map types */ << ";" << std::endl;
  }

  out << "input" << std::endl;
  for(std::set<exprt>::iterator it=invar.begin(); it!=invar.end(); it++) {
    out << *it << " : " /* map types */ << ";" << std::endl;
  }

  out << "local" << std::endl;
  for(std::map<exprt,exprt>::iterator it=lequ.begin(); it!=lequ.end(); it++) {
    out << it->first << " : " /* map types */ << ";" << std::endl;
  }

  out << "definition" << std::endl;
  for(std::map<exprt,exprt>::iterator it=lequ.begin(); it!=lequ.end(); it++) {
    out << it->first << " =" << it->second << ";" << std::endl;
  }

  out << "transition" << std::endl;
  for(std::map<exprt,exprt>::iterator it=sequ.begin(); it!=sequ.end(); it++) {
    out << it->first << "' =" << it->second << ";" << std::endl;
  }

  out << "initial " << INIT_VAR << ";" << std::endl;
  out << "assert true;" << std::endl; /* TODO: from input_assume */
  out << "invariant true;" << std::endl;

}

