/*******************************************************************\

Module: Test chain/vector generation base class

Author: Peter Schrammel

\*******************************************************************/

#ifndef CHAINCOVER_TEST_BASE_H
#define CHAINCOVER_TEST_BASE_H

#include <ostream>

#include <util/message.h>

//#include <ansi-c/ansi_c_language.h>
//#include "../goto-instrument/dump_c.h"

#include "testvector.h"
#include "bmc.h"
#include "tsmin/tsmin_template.h"

class test_baset : public messaget 
{
 public:
  test_baset(
    const optionst &_options,
    message_handlert &_message_handler):
    messaget(_message_handler),
    options(_options),
    ui(ui_message_handlert::PLAIN)
  {}

  typedef tsmin_templatet<unsigned,propertyt,unsigned> tsmin_chaincovert;

  virtual void compute(goto_functionst &goto_functions,
                       symbol_tablet &symbol_table) 
    { assert(false); }

  void minimize_test_suite()
  {
    tsmin_chaincovert::testsuite_desct tsdesc;
    tsmin_chaincovert::testgoalst goals;
    unsigned no = 0;
    for(testvectorst::const_iterator tv = testvectors.begin();
	tv != testvectors.end(); tv++, no++)
    {
      tsmin_chaincovert::testcase_desct tcdesc;
      tcdesc.id = no;
      for(propertiest::iterator p = tv->covered.begin();
	  p != tv->covered.end(); p++)
      {
	tcdesc.add_goal(*p);
	goals.insert(*p);
      }
      tcdesc.cost = tv->steps.size();
      tsdesc.add_testcase(tcdesc);
    }

    // minimize 
    tsmin_chaincovert::testsuitet ts_out;
    tsdesc.minimize_min_cost(ts_out,goals);

    // apply reduction
    testvectorst old_testvectors = testvectors;
    testvectors.clear();
    no = 0;
    for(testvectorst::const_iterator tv = old_testvectors.begin();
	tv != old_testvectors.end(); tv++, no++)
    {
      if(ts_out.find(no)!=ts_out.end()) 
	testvectors.push_back(*tv);
    }
  }

  void output_testvector(std::ostream &out, const testvectort &tv,
				 const namespacet &ns) 
  {
    out << "{";
    for(stepst::const_iterator v=tv.steps.begin();v!=tv.steps.end();) 
    {
      out << std::endl;
      for(valuationt::const_iterator vv=v->begin();vv!=v->end(); ) 
      {
        out << "  " << from_expr(ns,"",vv->first) << " = " 
	    << from_expr(ns,"",vv->second);
        if(++vv!=v->end()) out << ", ";
      }    
      if(++v!=tv.steps.end()) out << ", ";
    }
    out << std::endl << "}" << std::endl;
    out << "covers " << tv.covered.size() << " test goals:" << std::endl;
    out << "{" << std::endl;
    for(propertiest::const_iterator c=tv.covered.begin();
	c!=tv.covered.end();) 
    {
      out << "  " << *c; 
      if(++c!=tv.covered.end()) out << ",";
      out << std::endl;
    }
    out << "}" << std::endl << std::endl;
  }

  virtual void output_testvectors(std::ostream &out,
				  const namespacet &ns) 
  { 
    out << "\n" << testvectors.size() << " test vector(s) found: "  << std::endl;
    unsigned no = 0;
    for(testvectorst::const_iterator v = testvectors.begin(); 
        v!=testvectors.end(); v++, no++) 
    {
      out << "testvector" << no << "[" << v->steps.size() << "] = " <<std::endl;
      output_testvector(out,*v,ns);
    }
  }

  void xmloutput_testvector(std::ostream &out, const testvectort &tv,
				    const namespacet &ns) 
  {
    out << "    <covered-testgoals>" << std::endl;
    for(propertiest::const_iterator c=tv.covered.begin();
	c!=tv.covered.end(); ++c) 
    {
      out << "      <covered-testgoal id=\"" << *c << "\"/> " << std::endl;
    }
    out << "    </covered-testgoals>" << std::endl;
    out << "    <steps>" << std::endl;
    unsigned step = 0;
    for(stepst::const_iterator v=tv.steps.begin();
	v!=tv.steps.end(); ++v, ++step) 
    {
      out << "      <step number=\"" << step << "\">" << std::endl;
      for(valuationt::const_iterator vv=v->begin();vv!=v->end(); ++vv) 
      {
        out << "        <input lhs=\"" << from_expr(ns,"",vv->first) << 
               "\" value=\"" << from_expr(ns,"",vv->second) << 
                                       "\"/> " << std::endl;
      }    
      out << "      </step>" << std::endl;
    }
    out << "    </steps>" << std::endl;
  }

  void xmloutput_testvectors(std::ostream &out,
				     const namespacet &ns) 
  {
    out << "<testvectors>" << std::endl;
    unsigned no = 0;
    for(testvectorst::iterator v = testvectors.begin(); 
        v!=testvectors.end(); v++, no++) {
      out << "  <testvector number=\"" << no 
	  << "\" length=\"" << v->steps.size() << "\">" << std::endl;
      xmloutput_testvector(out,*v,ns);
      out << "  </testvector>" << std::endl;
    }
    out << "</testvectors>" << std::endl;
  }

  void output_tests_to_header_file(
     std::string filename, goto_functionst &goto_functions,
     const namespacet &ns) 
  {
    std::ofstream out(filename.c_str());
    if(!out.is_open()) 
    {
      error() << "Failed to write tests to header file '" << 
	filename << "'" << eom;
      return;
    }
    out << "#define TESTSUITE_SIZE " << testvectors.size() 
	<< std::endl  << std::endl;

    if(testvectors.size()==0) return;
    if(testvectors.front().steps.size()==0) return;
    if(testvectors.front().steps.front().size()!=1)
      warning() << "Some test inputs cannot be output" << eom;
    const typet &t = 
      ns.follow(testvectors.front().steps.front().begin()->first.type());

#if 0
    //TODO: There should be away to get the original struct type of the valuations
    out << "typedef " << from_type(ns,"",t) << " ";
    //TODO: should use dump_c:
    //goto2sourcet goto2c(goto_functions, ns, new_ansi_c_language);
    //goto2c.convert_compound(t,ns); //TODO: parsers clash
    if(t.id()==ID_struct)
    {
      const struct_union_typet &type = to_struct_union_type(t);
      out << "{" << std::endl;
      for(struct_union_typet::componentst::const_iterator
	    it=type.components().begin();
	  it!=type.components().end();
	  it++)
      {
	const struct_typet::componentt &comp=*it;
	typet comp_type=ns.follow(comp.type());

	irep_idt comp_name=comp.get_name();
	if(comp_type.id()==ID_code ||
	   comp.get_bool(ID_from_base) ||
	   comp.get_is_padding() ||
	   id2string(comp_name).find("$pad")!=std::string::npos)
	  continue;

	out << "  " << from_type(ns,"",comp_type) << " " 
	    << comp_name << ";" << std::endl;
      }      
      out << "}";
    }
    out << " __testvector_type;" << std::endl;
#endif 

    unsigned no = 0;
    for(testvectorst::const_iterator tv = testvectors.begin(); 
        tv!=testvectors.end();tv++,no++) 
    {
      out << "const unsigned testvector" << no 
	  << "_length = " << tv->steps.size() << ";" << std::endl;
      out << "const __testvector_type "
          << "testvector" << no << "[" << tv->steps.size() 
	  << "] = {" << std::endl;
      for(stepst::const_iterator v=tv->steps.begin();v!=tv->steps.end();) 
      {
        exprt e = v->begin()->second;
	remove_padding(e,ns);
        out << "  " << from_expr(ns,"",e);
        if(++v!=tv->steps.end()) out << ", " << std::endl;
      }
      out << std::endl << "};" << std::endl << std::endl;
    }
    out.close();
  }


  virtual std::string statistics_to_string() { return ""; }


  void add_testvector(const goto_functionst &goto_functions, bmct &bmc)
  {
    status() << "Building test vector..." << eom;
    bmc.goto_trace.clear();
    build_goto_trace_with_all_assertions(bmc.equation, bmc.symex.prop_conv, bmc.ns, bmc.goto_trace);
    testvectort testvector;
    //build chain
    for(goto_tracet::stepst::iterator it = bmc.goto_trace.steps.begin();
        it!=bmc.goto_trace.steps.end(); it++) 
    {
#if 0	
      std::cout << "code: " << from_expr(bmc.ns,"", it->pc->code) << std::endl;
#endif
      if(it->pc->is_assert() && !it->cond_value) 
	testvector.covered.insert(it->pc->location.get_comment());
      if(it->pc->is_function_call())
      {
	if(!it->full_lhs.is_nil())
	{
	  irep_idt function_name = 
            to_symbol_expr(to_code_function_call(
			   it->pc->code).function()).get_identifier();
	  if(goto_functions.function_map.find(function_name)==
	     goto_functions.function_map.end() ||
	     !goto_functions.function_map.at(function_name).body_available)
	  {
	    valuationt v;
	    v.insert(valuationt::value_type(it->full_lhs, it->full_lhs_value));
	    testvector.steps.push_back(v);
	  }
	}
      }
      if(it->pc->is_assign())
      {
	//TODO: main function parameters
	//TODO: uninitialized variables, etc
        if(to_code_assign(it->pc->code).rhs().id() == ID_sideeffect && 
           to_side_effect_expr(to_code_assign(
		 it->pc->code).rhs()).get_statement() == ID_nondet)
        {
          valuationt v;
          v.insert(valuationt::value_type(it->full_lhs, it->full_lhs_value));
          testvector.steps.push_back(v);
        }
      }
    }
    testvectors.push_back(testvector);
    status() << testvectors.size() << " test vectors" << eom;
  }

  testvectorst testvectors;

protected:
  const optionst &options;  
  language_uit::uit ui;
  
  struct goalt
  {
    exprt::operandst conjuncts;
    std::string description;
    bool covered; //goal reachable?

    explicit goalt(const goto_programt::instructiont &instruction)
    {
      description=id2string(instruction.location.get_comment());
      covered = false;
    }
  
    goalt() { }
  };

  // Collect _all_ goals in `goal_map'.
  // This maps claim IDs to 'goalt'
  typedef std::map<irep_idt, goalt> goal_mapt;
  goal_mapt goal_map;

  void remove_padding(exprt &expr, const namespacet &ns)
  {
    if(expr.type().id()==ID_struct)
    {
      exprt old_expr = expr;
      struct_union_typet &new_type = to_struct_union_type(expr.type());
      new_type.components().clear();
      expr.operands().clear();
      const struct_union_typet &old_type = to_struct_union_type(old_expr.type());
      for(unsigned i = 0; i<old_type.components().size(); i++)
      {
	const struct_typet::componentt &comp=old_type.components()[i];
	typet comp_type=ns.follow(comp.type());

	irep_idt comp_name=comp.get_name();
	if(id2string(comp_name).find("$pad")==std::string::npos)
	{
	  new_type.components().push_back(comp);
	  expr.operands().push_back(old_expr.operands()[i]);
	}
      }
    }
  }

};

#endif
