#include <iostream>

#include <goto-instrument/branchtest.h>
#include <util/xml.h>

#include "test_sat_base.h"

std::set<irep_idt> test_sat_baset::cover_test_goals(
    goto_functionst &goto_functions,
    bmct &bmc,
    symex_target_equationt::SSA_stepst::iterator equation_begin,
    const std::set<irep_idt> &covered_goals)
{
  prop_convt &prop_conv=bmc.symex.prop_conv;

  status() << "Passing problem to " << 
    prop_conv.decision_procedure_text() << eom;

  prop_conv.set_message_handler(get_message_handler());

  // stop the time
  absolute_timet sat_start=current_time();

  //convert  
  // bmc.equation.convert(prop_conv);
  bmc.equation.convert_guards(prop_conv);
  bmc.equation.convert_assignments(prop_conv);
  bmc.equation.convert_decls(prop_conv);
  bmc.equation.convert_assumptions(prop_conv);
  bmc.equation.convert_assertions(prop_conv);

  // Collect _all_ goals in `goal_map'.
  // This maps property IDs to 'goalt'
  goal_mapt goal_map;
  
  forall_goto_functions(f_it, goto_functions) {
    forall_goto_program_instructions(i_it, f_it->second.body) {
      if(i_it->is_assert()) {
        irep_idt property_id = i_it->location.get_property_id();
        if(covered_goals.find(property_id)==covered_goals.end())
	{ 
          goal_map[property_id]=goalt(*i_it);
	}
      }
    }
  }

  // get the conditions for these goals from formula
  unsigned property_counter=0;

  if(equation_begin==bmc.equation.SSA_steps.end()) 
    equation_begin=bmc.equation.SSA_steps.begin(); 
  for(symex_target_equationt::SSA_stepst::iterator
      it=equation_begin;
      it!=bmc.equation.SSA_steps.end();
      it++)
  {
    if(it->is_assert())
    {
      irep_idt property_id;

      if(it->source.pc->is_assert())
        property_id=it->source.pc->location.get_property_id();
      else
      {
        // need new property ID, say for an unwinding assertion
        property_counter++;
        property_id=i2string(property_counter);
        goal_map[property_id].description=it->comment;
      }

      if(covered_goals.find(property_id)==covered_goals.end()) 
        goal_map[property_id].conjuncts.push_back(
          literal_exprt(it->cond_literal));
    }
  }
  
  covert cover_goals(goto_functions,prop_conv,bmc,*this);

  //set activation literal for incremental checking
  bmc.equation.new_activation_literal(prop_conv);
  cover_goals.activation_literal = bmc.equation.current_activation_literal();
  
  for(goal_mapt::const_iterator
      it=goal_map.begin();
      it!=goal_map.end();
      it++)
  {
    // the following is TRUE if the bv is empty
    literalt p=!prop_conv.convert(conjunction(it->second.conjuncts));
    cover_goals.add(p);
  }

  status() << "Running "+prop_conv.decision_procedure_text() << eom;

  cover_goals();  

  // output runtime
  {
    std::ostringstream str;
    absolute_timet sat_stop=current_time();

    str << "Runtime decision procedure: ";
    str << (sat_stop-sat_start);
    str << "s";
    print(9,str.str());
  }
  
  // report
  if(ui!=ui_message_handlert::XML_UI)
  {
    status("");
    status("** Results:");
  }
  
  std::set<irep_idt> new_covered_goals;
  std::list<cover_goalst::cover_goalt>::const_iterator g_it=
    cover_goals.goals.begin();
  for(goal_mapt::const_iterator
      it=goal_map.begin();
      it!=goal_map.end();
      it++, g_it++)
  {
    if(ui==ui_message_handlert::XML_UI)
    {
      xmlt xml_result("result");
      xml_result.set_attribute("property", id2string(it->first));

      xml_result.set_attribute("status",
        g_it->covered?"FAILURE":"SUCCESS");
    
      std::cout << xml_result << std::endl;
    }
    else
    {
      status(std::string("[")+id2string(it->first)+"] "+
             it->second.description+": "+(g_it->covered?"COVERED":"NOT COVERED"));
      if(g_it->covered) 
      {
        new_covered_goals.insert(it->first);
      }
    }
  }

  status("");
  
  status("** "+i2string(cover_goals.number_covered())+
         " of "+i2string(cover_goals.size())+" covered ("+
         i2string(cover_goals.iterations())+" iterations)");
  return new_covered_goals;
}


void test_sat_baset::covert::assignment() 
{
  test.add_testvector(goto_functions,bmc);
}
