/*******************************************************************\

Module: SAT-based test case generation base class

Author: Peter Schrammel

\*******************************************************************/

#ifndef CHAINCOVER_TEST_SAT_BASE_H
#define CHAINCOVER_TEST_SAT_BASE_H

#include <map>

#include <goto-programs/goto_functions.h>
#include <util/options.h>
#include <langapi/language_ui.h>
#include <solvers/prop/cover_goals.h>
#include <util/time_stopping.h>

#include "../test_base.h"
#include "../bmc.h"

class test_sat_baset : public test_baset
{
public:
  test_sat_baset(
    const optionst &_options,
    message_handlert &_message_handler):
    test_baset(_options,_message_handler)
  {
  }
  
  virtual void compute(goto_functionst &goto_functions, 
                       symbol_tablet &symbol_table) 
    { assert(false); }

protected:
  
  struct goalt
  {
    exprt::operandst conjuncts;
    std::string description;
    bool covered; //goal reachable?

    explicit goalt(const goto_programt::instructiont &instruction)
    {
      description=id2string(instruction.location.get_comment());
      covered = false;
    }
  
    goalt() { }
  };


  // Collect _all_ goals in `goal_map'.
  // This maps claim IDs to 'goalt'
  typedef std::map<irep_idt, goalt> goal_mapt;



  class covert : public cover_goalst 
  {
   public:
  covert(goto_functionst &_goto_functions, prop_convt &_prop_conv, 
	 bmct& _bmc, test_sat_baset& _test):
    cover_goalst(_prop_conv),
      number_of_tests(0),goto_functions(_goto_functions),bmc(_bmc), test(_test) {}
  
    virtual void assignment();

    unsigned number_of_tests;

   protected:
    goto_functionst &goto_functions;
    bmct &bmc;
    test_sat_baset &test;
  };

  std::set<irep_idt> cover_test_goals(
    goto_functionst &goto_functions,
    bmct &bmc,
    symex_target_equationt::SSA_stepst::iterator equation_begin,
    const std::set<irep_idt> &covered_goals);
 
};


#endif
