#include <solvers/sat/satcheck.h>
#include <solvers/prop/literal_expr.h>
#include <goto-symex/slice.h>

#include "test_sat_incremental.h"
#include "../bmc.h"
#include "../cbmc_solvers.h"

void test_sat_incrementalt::compute(goto_functionst &goto_functions, 
                                    symbol_tablet &symbol_table)
{
  absolute_timet time_start(current_time());

  //get solver
  cbmc_solverst cbmc_solvers(options, symbol_table, get_message_handler());
  std::auto_ptr<cbmc_solverst::solvert> cbmc_solver = cbmc_solvers.get_solver();
  prop_convt& prop_conv = cbmc_solver->prop_conv();

  bmct bmc(options, symbol_table, get_message_handler(), prop_conv);

  //count goals
  unsigned number_of_goals = 0;
  for(goto_functionst::function_mapt::iterator f_it = 
          goto_functions.function_map.begin();
	  f_it!=goto_functions.function_map.end(); f_it++) 
  {
    for(goto_programt::targett i_it = f_it->second.body.instructions.begin();
	  i_it!=f_it->second.body.instructions.end(); i_it++) 
    {
      if(i_it->is_assert()) 
      {
	i_it->guard.make_false(); //activate testgoal assertion
        number_of_goals++;
      }
    }
  } 
  status() << "Number of test goals: " << number_of_goals << eom;

  //find entry point
  goto_functionst::function_mapt::const_iterator it=
      goto_functions.function_map.find(ID_main);
  if(it==goto_functions.function_map.end())
      throw "main symbol not found; please set an entry point";
  const goto_programt &goto_program=it->second.body;

  //setup incremental BMC
  bmc.symex.set_message_handler(get_message_handler());
  bmc.symex.options=options;
  bmc.symex.last_location.make_nil();
  bmc.setup_unwind();
 
  //incremental BMC loop
  goto_symext::statet symex_state;
  bool symex_done = false;
  symex_target_equationt::SSA_stepst::iterator equation_begin = 
    bmc.equation.SSA_steps.end();
  std::set<irep_idt> covered_goals;
  while(covered_goals.size()<number_of_goals && !symex_done) 
  { 
    symex_done = bmc.symex(symex_state,goto_functions,goto_program);
    undo_slice(bmc.equation); //undo all previous slicings
    debug() << "size of program expression: "
		 << bmc.equation.SSA_steps.size()
		 << " steps" << eom;
    slice(bmc.equation);
    std::set<irep_idt> new_covered_goals = 
      cover_test_goals(goto_functions,bmc,equation_begin,covered_goals);
    covered_goals.insert(new_covered_goals.begin(),new_covered_goals.end());

    debug() << "Number of covered goals: " << 
      covered_goals.size() << "/" << number_of_goals << eom;
    debug() << "time: " << (current_time()-time_start) << eom;
    equation_begin = bmc.equation.SSA_steps.end(); equation_begin--;
  }
  status() << "Number of test vectors: " << testvectors.size() << eom;
  status() << "Number of covered goals: " << 
      covered_goals.size() << "/" << number_of_goals << eom;
  statistics() << "Total time: " << (current_time()-time_start) << eom;
}

