#include <climits>

#include <util/prefix.h>
#include <goto-programs/set_properties.h>
#include <pointer-analysis/add_failed_symbols.h>

#include <goto-instrument/branchtest.h>

#include "test_sat_nonincremental.h"
#include "../bmc.h"
#include "../cbmc_solvers.h"

void test_sat_nonincrementalt::compute(goto_functionst &goto_functions, 
             symbol_tablet &symbol_table)
{
  namespacet ns(symbol_table);
  absolute_timet time_start(current_time());
  unsigned number_of_goals = 0;
  for(goto_functionst::function_mapt::iterator f_it = 
          goto_functions.function_map.begin();
	  f_it!=goto_functions.function_map.end(); f_it++) 
  {
    for(goto_programt::targett i_it = f_it->second.body.instructions.begin();
	  i_it!=f_it->second.body.instructions.end(); i_it++) 
    {
      if(i_it->is_assert()) 
      {
	i_it->guard.make_false(); //activate testgoal assertion
        number_of_goals++;
      }
    }
  } 
  status() << "Number of test goals: " << number_of_goals << eom;

  unsigned uncovered_goals = number_of_goals;
  unsigned max_unwind=options.get_unsigned_int_option("unwind");
  if(max_unwind==0) max_unwind = UINT_MAX;
  unsigned unwind = 1;
  while(uncovered_goals>0 && unwind<=max_unwind) 
  {
    //get solver
    cbmc_solverst cbmc_solvers(options, symbol_table, get_message_handler());
    std::auto_ptr<cbmc_solverst::solvert> cbmc_solver = 
      cbmc_solvers.get_solver();
    prop_convt& prop_conv = cbmc_solver->prop_conv();
    bmct bmc(options, symbol_table, get_message_handler(),prop_conv);
    bmc.symex.set_message_handler(get_message_handler());
    bmc.symex.options=options;
    bmc.symex.last_location.make_nil();
    bmc.setup_unwind();
    bmc.symex.set_unwind_limit(unwind); //set current unwind
    bmc.symex(goto_functions);

    std::set<irep_idt> covered_goals = 
      cover_test_goals(goto_functions,bmc,
                     bmc.equation.SSA_steps.begin(),
                     std::set<irep_idt>());

    if(!covered_goals.empty()) 
    {
      for(goto_functionst::function_mapt::iterator f_it = 
          goto_functions.function_map.begin();
	  f_it!=goto_functions.function_map.end(); f_it++) 
      {
        for(goto_programt::targett i_it = 
          f_it->second.body.instructions.begin();
	  i_it!=f_it->second.body.instructions.end(); i_it++) 
        {
          irep_idt property_id = i_it->location.get_property_id();
      	  if(covered_goals.find(property_id)!=covered_goals.end()) 
          {
	    i_it->make_skip();
	  }
	}
      } 
      uncovered_goals -= covered_goals.size();
      add_failed_symbols(symbol_table);
      goto_functions.update();
      goto_functions.compute_loop_numbers();
      label_properties(goto_functions);
    }
    unwind++;

    status() << "Number of test vectors: " << testvectors.size() << eom;
    status() << "Number of covered test goals: " << (number_of_goals-uncovered_goals) << "/" << number_of_goals << ", time: " << (current_time()-time_start) << eom;
  }
  statistics() << "Total time: " << (current_time()-time_start) << eom;
}
