#ifndef CHAINCOVER_CONFIG_PARSER_OPT_XML_H
#define CHAINCOVER_CONFIG_PARSER_OPT_XML_H

#include <fstream>
#include <iostream>
#include <util/message.h>
#include <goto-programs/goto_functions.h>
#include <util/std_expr.h>
#include <util/string2int.h>
#include <xmllang/xml_parser.h>

#include "reach_bmc.h"
#include "../config_parser_base.h"

#define DEFAULT_CHAIN_BOUND 0 //no bound, search optimum
#define DEFAULT_REACH_BOUND 10
#define DEFAULT_MAIN "c::main"
#define DEFAULT_INIT "c::init"
#define DEFAULT_FINAL "c::final"
#define DEFAULT_STEP "c::compute"
#define DEFAULT_STATE_ASSUME "c::state_assume"
#define DEFAULT_INPUT_ASSUME "c::input_assume"
#define DEFAULT_IO_TYPE "c::tag-io"
#define DEFAULT_STATE_TYPE "c::tag-state"


class config_parser_opt_xmlt : public config_parser_baset {
 public:
  config_parser_opt_xmlt(std::string _filename) : 
    config_parser_baset(_filename) {}

  virtual void parse() {
    null_message_handlert message_handler;
    xmlt xml;
    parse_xml(filename, message_handler, xml);
    for(xmlt::elementst::const_iterator it=xml.elements.begin();
        it!=xml.elements.end(); it++) {
      if(it->name=="chain") {
        for(xmlt::elementst::const_iterator itch=it->elements.begin();
            itch!=it->elements.end(); itch++) {
          if(itch->name=="init") init = itch->data;
          if(itch->name=="final") final = itch->data;
          if(itch->name=="reach-bound") reach_bound = 
            safe_string2unsigned(itch->data.c_str());
          if(itch->name=="chain-bound") chain_bound = 
            safe_string2unsigned(itch->data.c_str());
          if(itch->name=="testgoals") {
            for(xmlt::elementst::const_iterator itpr=itch->elements.begin();
                itpr!=itch->elements.end(); itpr++) {
              if(itpr->name=="testgoal") {
                properties.insert(itpr->data);
  	      }
            }
          }
	}
      }
    }
  }

  void parse(bmccontextt* context) {
    null_message_handlert message_handler;
    xmlt xml;
    parse_xml(filename, message_handler, xml);
    for(xmlt::elementst::const_iterator it=xml.elements.begin();
        it!=xml.elements.end(); it++) {
      if(it->name=="program") {
        for(xmlt::elementst::const_iterator itp=it->elements.begin();
            itp!=it->elements.end(); itp++) {
          if(itp->name=="state-type") context->stype = itp->data;
          if(itp->name=="io-type") context->itype = itp->data;
          if(itp->name=="state-assume") context->state_assume = itp->data;
          if(itp->name=="input-assume") context->input_assume = itp->data;
          if(itp->name=="init") context->init = itp->data;
          if(itp->name=="step") context->step = itp->data;
        }
      }
    }
  }

  virtual void infer(const goto_functionst& goto_functions,
		     const namespacet &ns) 
  {
    propertiest properties; 
    edgeweightt chain_bound = DEFAULT_CHAIN_BOUND;
    edgeweightt reach_bound = DEFAULT_REACH_BOUND;
    bmccontextt context;

    context.itype = DEFAULT_IO_TYPE;
    context.stype = DEFAULT_STATE_TYPE; 
    propertyt final = DEFAULT_FINAL; 
    context.init = DEFAULT_INIT; 
    context.step = DEFAULT_STEP; 
    context.state_assume = DEFAULT_STATE_ASSUME;
    context.input_assume = DEFAULT_INPUT_ASSUME;

    for(goto_functionst::function_mapt::const_iterator it = 
	  goto_functions.function_map.begin(); 
	it!=goto_functions.function_map.end(); it++) 
    {
      std::string fname = id2string(it->first);	
      if(fname.find("__CPROVER")!=std::string::npos) continue;
      //dumb matching on function names to detect a possible configuration
      if(fname.find("init")!=std::string::npos && it->second.type.arguments().size()>=1) {
        context.init = fname;
      }
      if(fname.find("final")!=std::string::npos && it->second.type.arguments().size()==2) 
        final = fname;
      if((fname.find("step")!=std::string::npos || fname.find("compute")!=std::string::npos) && 
	 it->second.type.arguments().size()==2) {
        context.step = fname; 
        dstring type1 = context.itype;
        dstring type2 = context.stype;
        //detect io and state types
	typet t1 = it->second.type.arguments()[0].type(); 
        if(t1.id()==ID_pointer ) {
          pointer_typet pt1 = to_pointer_type(t1);
          symbol_typet spt1 = to_symbol_type(pt1.subtype());
	  type1 = id2string(spt1.get_identifier());
	}
        typet t2 = it->second.type.arguments()[1].type();
        if(t2.id()==ID_pointer ) {
          pointer_typet pt2 = to_pointer_type(t2);
          symbol_typet spt2 = to_symbol_type(pt2.subtype());
	  type2 = id2string(spt2.get_identifier());
	}
        if(context.stype==type1) { //info from state-assume that state is first
          context.itype = type2; 
	}
        else { //no information: assume io is first, state is second
          context.itype = type1; 
          context.stype = type2; 
	}
      }
      if(fname.find("state_assume")!=std::string::npos && 
         it->second.type.arguments().size()==1){        
        //here, we have the chance to determine what is io and what state
        context.state_assume = fname;
        typet ts = it->second.type.arguments()[0].type();
        if(ts.id()==ID_pointer ) {
          pointer_typet pts = to_pointer_type(ts);
          symbol_typet spts = to_symbol_type(pts.subtype());
          
	  dstring stype = id2string(spts.get_identifier());
	  if(context.itype==stype) { //swap if we guessed them wrong
            context.itype = context.stype;
	  }
          context.stype = stype;
	}
      }
      if(fname.find("input_assume")!=std::string::npos && 
         it->second.type.arguments().size()==2) {        
        context.input_assume = fname;
      }
      if(fname.find("prop")!=std::string::npos && it->second.type.arguments().size()==2) 
        properties.insert(fname);
    }
    propertyt init = context.init; //same as init() by default

    //write to file
    std::ofstream outfile(filename.c_str());
    if (!outfile.is_open()) throw ("ERROR: cannot open file "+filename);
    outfile << "<chaincover-config>" << std::endl;
    outfile << "  <program>" << std::endl;
    outfile << "    <state-type>" << context.stype << "</state-type>" << std::endl;
    outfile << "    <io-type>" << context.itype << "</io-type>" << std::endl;
    outfile << "    <state-assume>" << context.state_assume << "</state-assume>" << std::endl;
    outfile << "    <input-assume>" << context.input_assume << "</input-assume>" << std::endl;
    outfile << "    <init>" << context.init << "</init>" << std::endl;
    outfile << "    <step>" << context.step << "</step>" << std::endl;
    outfile << "  </program>" << std::endl;
    outfile << "  <chain>" << std::endl;
    outfile << "    <init>" << init << "</init>" << std::endl;
    outfile << "    <final>" << final << "</final>" << std::endl;
    outfile << "    <reach-bound>" << reach_bound << "</reach-bound>" << std::endl;
    outfile << "    <chain-bound>" << chain_bound << "</chain-bound>" << std::endl;
    outfile << "    <testgoals>" << std::endl;
    for(propertiest::iterator p = properties.begin(); p!=properties.end(); p++) {
      outfile << "      <testgoal>" << *p << "</testgoal>" << std::endl;
    }
    outfile << "    </testgoals>" << std::endl;
    outfile << "  </chain>" << std::endl;
    outfile << "</chaincover-config>" << std::endl;
    outfile.close();
  }

  propertyt init;
  edgeweightt reach_bound;

};

#endif
