#include "hampath_solver_lkh.h"
#include "LKH.h"
#include "Heap.h"
#include "ReadProblem.h"
#include <iostream>

template<class PropT>
typename graph_weightedt<nodeidt, PropT, edgeidt>::patht hampath_solver_lkht<PropT>::solve(graph_weightedt<nodeidt, PropT, edgeidt>* graph, nodeidt init, nodeidt final, edgeweightt bound)
/*
  typename hampath_solver_lkht<PropT>::patht hampath_solver_lkht<PropT>::solve(hampath_solver_lkht<PropT>::grapht* graph, nodeidt init, nodeidt final, edgeweightt bound)*/
{
    GainType Cost;
    double Time, LastTime = GetTime();

    /* Read the specification of the problem */
    setParameters(bound);
    MaxMatrixDimension = graph->getNodeCount()+1;
    setProblem(graph,init,final);

    AllocateStructures();
    CreateCandidateSet();
    InitializeStatistics();

    if (Norm != 0)
        BestCost = PLUS_INFINITY;
    else {
        /* The ascent has solved the problem! */
        Optimum = BestCost = (GainType) LowerBound;
        UpdateStatistics(Optimum, GetTime() - LastTime);
        RecordBetterTour();
        RecordBestTour();
        WriteTour(OutputTourFileName, BestTour, BestCost);
        WriteTour(TourFileName, BestTour, BestCost);
        Runs = 0;
    }

    /* Find a specified number (Runs) of local optima */
    for (Run = 1; Run <= Runs; Run++) {
        LastTime = GetTime();
        Cost = FindTour();      /* using the Lin-Kernighan heuristic */
        if (Run > 1) {
            Cost = MergeBetterTourWithBestTour();
            if (Cost < BestCost)
                RecordBetterTour();
        }
        if (Cost < BestCost) {
            BestCost = Cost;
            RecordBestTour();
            if (MaxTrials == 0)
                WriteTour(OutputTourFileName, BestTour, BestCost);
            WriteTour(TourFileName, BestTour, BestCost);
        }
        if (Cost < Optimum) {
            if (FirstNode->InputSuc) {
                Node *N = FirstNode;
                while ((N = N->InputSuc = N->Suc) != FirstNode);
            }
            Optimum = Cost;
	    std::cout << "*** New optimum = " << Optimum << " ***" << std::endl << std::endl;
	    //            printff("*** New optimum = " GainFormat " ***\n\n", Optimum);
        }
        Time = fabs(GetTime() - LastTime);
        UpdateStatistics(Cost, Time);
        if (TraceLevel >= 1 && Cost != PLUS_INFINITY) {
	    std::cout << "Run " << Run << ": Cost = " << Cost;
	    //            printff("Run %d: Cost = " GainFormat, Run, Cost);
            if (Optimum != MINUS_INFINITY && Optimum != 0)
	      std::cout << ", Gap = " << (100.0 * (Cost - Optimum) / Optimum);
            std::cout << ", Time = " << Time << " sec. " << (Cost < Optimum ? "<" : Cost == Optimum ? "=" : "") << std::endl << std::endl;
		  /*            printff(", Time = %0.1f sec. %s\n\n", Time,
				Cost < Optimum ? "<" : Cost == Optimum ? "=" : "");*/
        }
        SRandom(++Seed);
    }
    PrintStatistics();

  //extract path from init to final
  if(BestCost>=WEIGHT_UNCONNECTED) return typename hampath_solver_lkht<PropT>::patht();//no path found
  int nn = ProblemType != ATSP ? Dimension : Dimension / 2;
  typename hampath_solver_lkht<PropT>::patht p;
  int i=0;
  std::cout << "best tour: ";
  for(i=0;i<nn;i++) std::cout << BestTour[i] << " ";
  std::cout << std::endl;
  i=0;
  while(i<nn && (BestTour[i]-1)!=init) { i++; } //search init
  if(i==nn) throw "ERROR: no init node in TSP path";
  if(BestTour[(i-1)%nn>=0 ? (i-1)%nn : (nn+i-1)%nn]-1!=final) 
    throw "ERROR: final and init nodes not adjacent in TSP path";
  while(BestTour[i%nn]-1!=final) { 
    p.push_back(BestTour[i%nn]-1);
    i++;
  }
  p.push_back(BestTour[i%nn]-1); //push final
  return p;
}

template<class PropT>
void hampath_solver_lkht<PropT>::setParameters(edgeweightt bound) {
    ProblemFileName = PiFileName = InputTourFileName =
        OutputTourFileName = TourFileName = 0;
    CandidateFiles = MergeTourFiles = 0;
    AscentCandidates = 50;
    BackboneTrials = 0;
    Backtracking = 0;
    CandidateSetSymmetric = 0;
    CandidateSetType = ALPHA;
    DelaunayPartitioning = 0;
    DelaunayPure = 0;
    Excess = -1;
    ExtraCandidates = 0;
    ExtraCandidateSetSymmetric = 0;
    ExtraCandidateSetType = QUADRANT;
    Gain23Used = 1;
    GainCriterionUsed = 1;
    InitialPeriod = -1;
    InitialStepSize = 0;
    InitialTourAlgorithm = WALK;
    InitialTourFraction = 1.0;
    KarpPartitioning = 0;
    KMeansPartitioning = 0;
    Kicks = 1;
    KickType = 0;
    MaxBreadth = INT_MAX;
    MaxCandidates = 5;
    MaxSwaps = -1;
    MaxTrials = -1;
    MoorePartitioning = 0;
    MoveType = 5;
    NonsequentialMoveType = -1;
    Optimum = bound; //MINUS_INFINITY;
    PatchingA = 2; //1;
    PatchingC = 3; ;//0;
    PatchingAExtended = 0;
    PatchingARestricted = 0;
    PatchingCExtended = 0;
    PatchingCRestricted = 0;
    Precision = 1; //100;
    RestrictedSearch = 1;
    RohePartitioning = 0;
    Runs = 10; //0;
    Seed = 1;
    SierpinskiPartitioning = 0;
    StopAtOptimum = 1; //stop if tour <=bound is found
    Subgradient = 1;
    SubproblemsCompressed = 0;
    SubproblemSize = 0;
    SubsequentMoveType = 0;
    SubsequentPatching = 1;
    TimeLimit = DBL_MAX;
    TraceLevel = 1;
}

template<class PropT>
void hampath_solver_lkht<PropT>::setProblem(grapht* graph, nodeidt init, nodeidt final)
{
    int i, K;
    //    char *Line, *Keyword;

    /*        if (!(ProblemFile = fopen(ProblemFileName, "r")))
        eprintf("Cannot open PROBLEM_FILE: \"%s\"", ProblemFileName);
    if (TraceLevel >= 1)
    printff("Reading PROBLEM_FILE: \"%s\" ... ", ProblemFileName);*/
    FreeStructures();
    FirstNode = 0;
    ProblemType = ATSP;
    WeightType = EXPLICIT;
    Distance = Distance_EXPLICIT;
    WeightFormat = FULL_MATRIX;
    DimensionSaved = Dimension = graph->getNodeCount();
    CoordType = NO_COORDS;
    //Name = "TestChain";
    Type = EdgeWeightType = EdgeWeightFormat = 0;
    EdgeDataFormat = NodeCoordType = DisplayDataType = 0;
    Distance = 0;
    C = 0;
    c = 0;
    /*   while ((Line = ReadLine(ProblemFile))) {
        if (!(Keyword = strtok(Line, Delimiters)))
            continue;
        for (i = 0; i < (int) strlen(Keyword); i++)
            Keyword[i] = (char) toupper(Keyword[i]);
        if (!strcmp(Keyword, "COMMENT"));
        else if (!strcmp(Keyword, "DEMAND_SECTION"))
            eprintf("Not implemented: %s", Keyword);
        else if (!strcmp(Keyword, "DEPOT_SECTION"))
            eprintf("Not implemented: %s", Keyword);
        else if (!strcmp(Keyword, "DIMENSION"))
            Read_DIMENSION();
        else if (!strcmp(Keyword, "DISPLAY_DATA_SECTION"))
            Read_DISPLAY_DATA_SECTION();
        else if (!strcmp(Keyword, "DISPLAY_DATA_TYPE"))
            Read_DISPLAY_DATA_TYPE();
        else if (!strcmp(Keyword, "EDGE_DATA_FORMAT"))
            Read_EDGE_DATA_FORMAT();
        else if (!strcmp(Keyword, "EDGE_DATA_SECTION"))
            Read_EDGE_DATA_SECTION();
        else if (!strcmp(Keyword, "EDGE_WEIGHT_FORMAT"))
            Read_EDGE_WEIGHT_FORMAT();
        else if (!strcmp(Keyword, "EDGE_WEIGHT_SECTION"))
            Read_EDGE_WEIGHT_SECTION();
        else if (!strcmp(Keyword, "EDGE_WEIGHT_TYPE"))
            Read_EDGE_WEIGHT_TYPE();
        else if (!strcmp(Keyword, "EOF"))
            break;
        else if (!strcmp(Keyword, "FIXED_EDGES_SECTION"))
            Read_FIXED_EDGES_SECTION();
        else if (!strcmp(Keyword, "NAME"))
            Read_NAME();
        else if (!strcmp(Keyword, "NODE_COORD_SECTION"))
            Read_NODE_COORD_SECTION();
        else if (!strcmp(Keyword, "NODE_COORD_TYPE"))
            Read_NODE_COORD_TYPE();
        else if (!strcmp(Keyword, "TOUR_SECTION"))
            Read_TOUR_SECTION(&ProblemFile);
        else if (!strcmp(Keyword, "TYPE"))
            Read_TYPE();
        else
            eprintf("Unknown keyword: %s", Keyword);
	    }*/

    Node *Ni, *Nj;
    int j, n;
    GainType W;

    CheckSpecificationPart();
    if (!FirstNode)
        CreateNodes();
    if (ProblemType != ATSP) {
        assert(CostMatrix =
               (int *) calloc(Dimension * (Dimension - 1) / 2,
                              sizeof(int)));
        Ni = FirstNode->Suc;
        do {
            Ni->C = &CostMatrix[(Ni->Id - 1) * (Ni->Id - 2) / 2] - 1;
        }
        while ((Ni = Ni->Suc) != FirstNode);
    } else {
        n = Dimension / 2;
        assert(CostMatrix = (int *) calloc(n * n, sizeof(int)));
        for (Ni = FirstNode; Ni->Id <= n; Ni = Ni->Suc)
            Ni->C = &CostMatrix[(Ni->Id - 1) * n] - 1;
    }       
    if (ProblemType == ATSP) {
            int n = Dimension / 2;
            for (i = 1; i <= n; i++) {
                Ni = &NodeSet[i];
                for (j = 1; j <= n; j++) {
                  //get edge weights
		  typename hampath_solver_lkht<PropT>::grapht::edget edge = typename hampath_solver_lkht<PropT>::grapht::edget(i-1,j-1);
                  if(i==j) W = 0;
                  else if (i-1==final && j-1==init) W = 0;
                  else if(!graph->existsEdge(edge)) W = WEIGHT_UNCONNECTED;
		  else W = graph->getEdgeLabel(edge);
		  std::cout << "weight for (" << i << "," << j << ") = " << W << std::endl;
		  /*if (!fscanint(ProblemFile, &W))
		    eprintf("Missing weight in EDGE_WEIGHT_SECTION");*/
                    Ni->C[j] = W;
                    if (i != j && W > M)
                        M = W;
                }
                Nj = &NodeSet[i + n];
                if (!Ni->FixedTo1)
                    Ni->FixedTo1 = Nj;
                else if (!Ni->FixedTo2)
                    Ni->FixedTo2 = Nj;
                if (!Nj->FixedTo1)
                    Nj->FixedTo1 = Ni;
                else if (!Nj->FixedTo2)
                    Nj->FixedTo2 = Ni;
            }
            Distance = Distance_ATSP;
            WeightType = -1;
        } else
            for (i = 1, Ni = FirstNode; i <= Dimension; i++, Ni = Ni->Suc) {
                for (j = 1; j <= Dimension; j++) {
                  //get edge weights
		  typename hampath_solver_lkht<PropT>::grapht::edget edge = typename hampath_solver_lkht<PropT>::grapht::edget(i-1,j-1);
                  if(i==j) W = 0;
                  else if (i-1==final && j-1==init) W = 0;
                  else if(!graph->existsEdge(edge)) W = WEIGHT_UNCONNECTED;
		  else W = graph->getEdgeLabel(edge);
		  /*if (!fscanint(ProblemFile, &W))
		    eprintf("Missing weight in EDGE_WEIGHT_SECTION");*/
                    if (j < i)
                        Ni->C[j] = W;
                }
            }

    Swaps = 0;

    /* Adjust parameters */
    if (Seed == 0)
        Seed = 1;
    if (Precision == 0)
        Precision = 100;
    if (InitialStepSize == 0)
        InitialStepSize = 1;
    if (MaxSwaps < 0)
        MaxSwaps = Dimension;
    if (KickType > Dimension / 2)
        KickType = Dimension / 2;
    if (Runs == 0)
        Runs = 10;
    if (MaxCandidates > Dimension - 1)
        MaxCandidates = Dimension - 1;
    if (ExtraCandidates > Dimension - 1)
        ExtraCandidates = Dimension - 1;
    if (SubproblemSize >= Dimension)
        SubproblemSize = Dimension;
    else if (SubproblemSize == 0) {
        if (AscentCandidates > Dimension - 1)
            AscentCandidates = Dimension - 1;
        if (InitialPeriod < 0) {
            InitialPeriod = Dimension / 2;
            if (InitialPeriod < 100)
                InitialPeriod = 100;
        }
        if (Excess < 0)
            Excess = 1.0 / Dimension;
        if (MaxTrials == -1)
            MaxTrials = Dimension;
        MakeHeap(Dimension);
    }

    if (CostMatrix == 0 && Dimension <= MaxMatrixDimension && Distance != 0
        && Distance != Distance_1 && Distance != Distance_ATSP &&
        WeightType != GEO && WeightType != GEOM &&
        WeightType != GEO_MEEUS && WeightType != GEOM_MEEUS) {
        Node *Ni, *Nj;
        assert(CostMatrix =
               (int *) calloc(Dimension * (Dimension - 1) / 2,
                              sizeof(int)));
        Ni = FirstNode->Suc;
        do {
            Ni->C = &CostMatrix[(Ni->Id - 1) * (Ni->Id - 2) / 2] - 1;
            if (ProblemType != HPP || Ni->Id < Dimension)
                for (Nj = FirstNode; Nj != Ni; Nj = Nj->Suc)
                    Ni->C[Nj->Id] = Fixed(Ni, Nj) ? 0 : Distance(Ni, Nj);
            else
                for (Nj = FirstNode; Nj != Ni; Nj = Nj->Suc)
                    Ni->C[Nj->Id] = 0;
        }
        while ((Ni = Ni->Suc) != FirstNode);
        WeightType = EXPLICIT;
        c = 0;
    }
    if (Precision > 1 && (WeightType == EXPLICIT || ProblemType == ATSP)) {
        int i, j, n = ProblemType == ATSP ? Dimension / 2 : Dimension;
        for (i = 2; i <= n; i++) {
            Node *N = &NodeSet[i];
            for (j = 1; j < i; j++)
                if (N->C[j] * Precision / Precision != N->C[j])
                    eprintf("PRECISION (= %d) is too large", Precision);
        }
    }
    C = WeightType == EXPLICIT ? C_EXPLICIT : C_FUNCTION;
    D = WeightType == EXPLICIT ? D_EXPLICIT : D_FUNCTION;
    if (SubsequentMoveType == 0)
        SubsequentMoveType = MoveType;
    K = MoveType >= SubsequentMoveType
        || !SubsequentPatching ? MoveType : SubsequentMoveType;
    if (PatchingC > K)
        PatchingC = K;
    if (PatchingA > 1 && PatchingA >= PatchingC)
        PatchingA = PatchingC > 2 ? PatchingC - 1 : 1;
    if (NonsequentialMoveType == -1 ||
        NonsequentialMoveType > K + PatchingC + PatchingA - 1)
        NonsequentialMoveType = K + PatchingC + PatchingA - 1;
    if (PatchingC >= 1 && NonsequentialMoveType >= 4) {
        BestMove = BestSubsequentMove = BestKOptMove;
        if (!SubsequentPatching && SubsequentMoveType <= 5) {
            MoveFunction BestOptMove[] =
                { 0, 0, Best2OptMove, Best3OptMove,
                Best4OptMove, Best5OptMove
            };
            BestSubsequentMove = BestOptMove[SubsequentMoveType];
        }
    } else {
        MoveFunction BestOptMove[] = { 0, 0, Best2OptMove, Best3OptMove,
            Best4OptMove, Best5OptMove
        };
        BestMove = MoveType <= 5 ? BestOptMove[MoveType] : BestKOptMove;
        BestSubsequentMove = SubsequentMoveType <= 5 ?
            BestOptMove[SubsequentMoveType] : BestKOptMove;
    }
    if (TraceLevel >= 1) {
      std::cout << "done" << std::endl;
      //printff("done\n");
        PrintParameters();
    } else
      /*    printff("PROBLEM_FILE = %s\n",
	    ProblemFileName ? ProblemFileName : "");*/
    fclose(ProblemFile);
    if (InitialTourFileName)
        ReadTour(InitialTourFileName, &InitialTourFile);
    if (InputTourFileName)
        ReadTour(InputTourFileName, &InputTourFile);
    if (SubproblemTourFileName && SubproblemSize > 0)
        ReadTour(SubproblemTourFileName, &SubproblemTourFile);
    if (MergeTourFiles >= 1) {
        free(MergeTourFile);
        assert(MergeTourFile =
               (FILE **) malloc(MergeTourFiles * sizeof(FILE *)));
        for (i = 0; i < MergeTourFiles; i++)
            ReadTour(MergeTourFileName[i], &MergeTourFile[i]);
    }
    free(LastLine);
    LastLine = 0;
}
