/*******************************************************************\

Module: BMC-based reachability engine

Author: Peter Schrammel

\*******************************************************************/

#ifndef CHAINCOVER_REACH_BMC_H
#define CHAINCOVER_REACH_BMC_H
 
#include "../testvector.h"
#include "reach_template.h"
#include <util/options.h>
#include <langapi/language_ui.h>
#include "../symex_bmc.h"

/*
#include <mutex>
#define NUM_THREADS 2
#define MULTITHREADING_K 2
*/

class bmccontextt {
 public:
  const optionst* options;
  language_uit::uit ui;
  message_handlert* ui_message_handler;
  const goto_functionst* goto_functions;
  symbol_tablet* symbol_table;
  irep_idt itype; // e.g. "c::tag-input"
  irep_idt stype; // e.g. "c::tag-state"
  irep_idt state_assume; // e.g. "c::state_assume"
  irep_idt input_assume; // e.g. "c::input_assume"
  irep_idt init; // e.g. "c::init"
  irep_idt step; // e.g. "c::compute"
  std::string harness_incfile;
};

#include "../bmc.h"

class reach_bmct : public reach_templatet<bmccontextt,propertyt,testvectort>, public messaget {
 public:
 reach_bmct(bmccontextt* _context) :  
      messaget(*_context->ui_message_handler),
      context(_context)  {
  }

  /** checks k-reachability from prop1 to prop2 */
  virtual resultt check_k_reach(propertyt prop1, propertyt prop2, unsigned k);
  virtual prop_edgest check_k_reach(propertyt prop1, propertiest props, unsigned k);

  /** computes the set of k-step edges */
  virtual prop_edgest get_k_reach_edges(prop_edgest edges, unsigned k);

  /** computes the test case along the given path (not used) */
  virtual bool get_chain(testvectort& chain, prop_patht& failed_path, prop_patht path, pathweightst pathweightse); 

  /** checks a given path */
  virtual bool check_path(prop_patht path, pathweightst pathweights);

  /** chain repair */
  virtual void repair_chain(prop_patht proppath, pathweightst& pathweights, prop_patht& failed_path, edgeweightt reach_bound, testvectort &chain);

  /** prints a chain */
  virtual std::ostream &output_chain(testvectort chain, std::ostream& out) {
    namespacet ns(*context->symbol_table);
    for(stepst::iterator v=chain.steps.begin();v!=chain.steps.end();) {
      out << std::endl;
      for(valuationt::iterator vv=v->begin();vv!=v->end(); ) {
        out << "  " << from_expr(ns,"",vv->first) << " = " 
	    << from_expr(ns,"",vv->second);
         if(++vv!=v->end()) out << ", ";
      }    
      if(++v!=chain.steps.end()) out << ", ";
    }
    return out;
  }

  virtual std::ostream &xmloutput_chain(testvectort chain, std::ostream& out) {
    namespacet ns(*context->symbol_table);
    for(stepst::iterator v=chain.steps.begin();v!=chain.steps.end(); ++v) {
      out << "      <step>" << std::endl;
      for(valuationt::iterator vv=v->begin();vv!=v->end(); ++vv) {
        out << "        <input lhs=\"" << from_expr(ns,"",vv->first) << 
               "\" value=\"" << from_expr(ns,"",vv->second) << 
                                       "\"/> " << std::endl;
      }    
      out << "      </step>" << std::endl;
    }
    return out;
  }

  virtual void output_chain_file(testvectort chain, std::string outfile);

  std::ostream &output_context(std::ostream& out) {
    out << "BMC reachability engine context:" << std::endl;
    out << "  io-type = " << context->itype << std::endl;
    out << "  state-type = " << context->stype << std::endl;
    out << "  state-assume = " << context->state_assume << std::endl;
    out << "  init = " << context->init << std::endl;
    out << "  step = " << context->step << std::endl;
    return out;
  }

  std::string context_to_string() {
    std::ostringstream out; 
    output_context(out);
    return out.str();
  }

  /**** experimental stuff */
  virtual void output_harness(testvectort chain, prop_patht path, pathweightst pathweights, std::string outfile);

  //  void output_goto_nbac(irep_idt init, irep_idt step, irep_idt input_assume, std::ostream& out);
  /************/

 protected:
  bmccontextt* context;

  void add_prop_entry(goto_programt &goto_program,irep_idt prop_name,bool as_assertion,symbol_exprt ivar, symbol_exprt svar);
  void add_function_call(goto_programt &goto_program,const irep_idt name,exprt arg1, irep_idt argtype1, exprt arg2, irep_idt argtype2);
  void add_var_declaration(goto_programt &goto_program,symbol_exprt& var, irep_idt name, irep_idt pretty_name, irep_idt base_name, typet type);
  void add_nondet_assign_inputs(goto_programt &goto_program,exprt iexpr, exprt iiexpr);
  void add_assign_state(goto_programt &goto_program, symbol_exprt ssexpr, valuationt v);
  bool check_steps(propertyt beginprop, valuationt beginstate, propertyt endprop, unsigned steps, valuationt& endstate, testvectort &chain);
  bool check_path_chain(prop_patht path, pathweightst pathweights, valuationt& endstate, testvectort& chain); 

  prop_edgest get_k_reach_edges_sequential(prop_edgest edges, unsigned k);
  prop_edgest get_k_reach_edges_cover(prop_edgest edges, unsigned k);
  //  prop_edgest get_k_reach_edges_parallel(prop_edgest edges, unsigned k);

};

#endif
