/*******************************************************************\

Module: Reachability engine interface

Author: Peter Schrammel

\*******************************************************************/

#ifndef CHAINCOVER_REACH_TEMPLATE_H
#define CHAINCOVER_REACH_TEMPLATE_H

#include "graph_weighted.h"
#include <set>
#include <iostream>

template<class ContextT, class PropT, class ChainT>
class reach_templatet {
 public:
  typedef enum{REACH, UNREACH, ERROR} resultt;
  typedef std::pair<PropT,PropT> prop_edget;
  typedef std::set<prop_edget> prop_edgest;
  typedef std::list<PropT> prop_patht;
  typedef std::list<PropT> propertiest;
  typedef std::list<edgeweightt> pathweightst;

  void set_context(ContextT _context) {
    context = _context;
  }

  // returns the set of edges from prop1 to props satisfying k-reachability
  virtual prop_edgest check_k_reach(PropT prop1, propertiest props, edgeweightt k) = 0;

  // computes the subset of edges satisfying k-reachability
  virtual prop_edgest get_k_reach_edges(prop_edgest edges, unsigned k) = 0;

  // checks the feasibility of the given path
  virtual bool check_path(prop_patht path, pathweightst pathweights) = 0;

  // returns an empty chain if reach_bound exceeded, failed_path contains the failed 3-subpath
  virtual void repair_chain(prop_patht proppath, pathweightst& pathweights, prop_patht& failed_path, edgeweightt reach_bound, ChainT &chain) = 0;

  virtual resultt check_k_reach(PropT prop1, PropT prop2, edgeweightt k) = 0; // (not used)
  virtual bool get_chain(ChainT& chain, prop_patht& failed_path, prop_patht path, pathweightst pathweights) = 0; // (not used)

  //printing
  virtual std::ostream &output_chain(ChainT chain, std::ostream& out) = 0;
  virtual std::ostream &xmloutput_chain(ChainT chain, std::ostream& out) = 0;

  std::ostream &output_path(prop_patht path, pathweightst pathweights, std::ostream& out) {
    out << "{ ";
    typename prop_patht::iterator p = path.begin();
    typename pathweightst::iterator w = pathweights.begin();
    for(; p != path.end(); p++) {
      out << *p;
      if(w != pathweights.end()) {
   	out << " -- " << *w << " --> ";
        w++;
      } 
    }
    out << " }";
    return out;
  }

  // * experimental *
  virtual void output_harness(ChainT chain, prop_patht path, pathweightst pathweights, std::string outfile) = 0;


 protected:
  ContextT context;
};

#endif
