#include "testchain_opt.h"

#include "hampath_solver_lkh.h"
#include "hampath_solver_clingo.h"
#include "reach_bmc.h"
#include "chain.h"

void testchain_optt::compute(goto_functionst &goto_functions,
                       symbol_tablet &symbol_table)
{
  bmccontextt context;
  context.options = &options;
  context.ui_message_handler = message_handler;
  context.ui = ui;
  context.goto_functions = &goto_functions;
  context.symbol_table = &symbol_table;
  //context.harness_incfile = cmdline.args[0]; // *experimental*

  //hamiltonian path solver
  hampath_solvert* hampath_solver;
  if(options.get_bool_option("clingo")) 
  {
    hampath_solver = new hampath_solver_clingot<propertyt>(
			  *message_handler);
  }
  else 
  {
    hampath_solver = new hampath_solver_lkht<propertyt>(
                          *message_handler);
  }

  config->parse(&context);

  //reachability query solver
  reach_bmct reach_bmc(&context);
  debug(reach_bmc.context_to_string());

  //configure test chain generator
  testchaint testchain(&reach_bmc,hampath_solver,config,
                       *message_handler);
  testchain.read_config();
  debug(testchain.config_to_string());  
  //testchain.set_harness_outfile(cmdline.getval("output-harness")); // *experimental*
  testchain.set_use_path_constraints(
    options.get_bool_option("path-constraints"));
  testchain.set_compute_multichain(!options.get_bool_option("no-multi-chain"));

  //compute chain
  testvectors = testchain.compute_chain();    

  statistics(testchain.statistics_to_string());

#if 0
  if(cmdline.isset("output-chain")) {
    if(chains.size()!=1) throw "Can only output a single chain";
    reach_bmc.output_chain_file(chains.front(),cmdline.getval("output-chain"));
    status(std::string("chain written to ")+cmdline.getval("output-chain"));
  }
#endif
  
  delete hampath_solver;
}
