#include <analyses/natural_loops.h>

#include <goto-programs/goto_functions.h>

#include <util/std_expr.h>

#include <iostream>
#include <list>

#include "accelerate.h"
#include "util.h"
#include "path.h"
#include "polynomial_accelerator.h"
#include "looppath_select_all.h"
#include "looppath_select_filter.h"

//#define DEBUG

int acceleratet::accelerate_loop(goto_programt::targett loop_header) {
  pathst loop_paths;
  int num_accelerated = 0;

  goto_programt::instructiont skip(SKIP);
  program.insert_before_swap(loop_header, skip);
  program.compute_incoming_edges();

  goto_programt::targett new_inst = loop_header;
  ++new_inst;

  natural_loops_mutablet::natural_loopt loop = 
    natural_loops.loop_map.find(loop_header)->second;
  loop.insert(new_inst);

  looppath_select.find_paths(loop_header, loop, loop_paths);
  std::cout << loop_paths.size() << " path(s) found" << std::endl;

  for (pathst::iterator it = loop_paths.begin();
       it != loop_paths.end();
       ++it) {
    goto_programt accelerator;

    if (accelerate(*it, accelerator)) {
      add_accelerator(loop_header, accelerator);
      num_accelerated++;
    }
  }

  return num_accelerated;
}

void acceleratet::add_accelerator(goto_programt::targett loop_header,
                                  goto_programt &accelerator) {
  //find end of loop
  targetst preds = loop_header->incoming_edges;
  assert(preds.size()>0);
  goto_programt::targett loop_end = *(preds.begin());
  for(targetst::iterator it = preds.begin(); it!=preds.end(); it++) {
    if((*it)->location_number > loop_header->location_number) {
      loop_end = *it; 
      break;
    }
  }
    
  goto_programt::targett loop_body = loop_header;
  ++loop_body;

  goto_programt::targett jump = program.insert_before(loop_body);
  jump->make_goto();
  jump->guard = side_effect_expr_nondett(bool_typet());
  jump->targets.push_back(loop_body);

  for(goto_programt::targett it = accelerator.instructions.begin(); 
      it != accelerator.instructions.end(); it++) {
    program.instructions.insert(loop_body,*it);
  }

  jump = program.insert_before(loop_body);
  jump->make_goto();
  jump->guard = true_exprt();
  jump->targets.push_back(loop_end); //jump to end of loop (not to header!)
  jump->location.set_comment("accelerator");
}

int acceleratet::accelerate_loops()
{
  int num_accelerated = 0;

  for (natural_loops_mutablet::loop_mapt::iterator it =
          natural_loops.loop_map.begin();
       it != natural_loops.loop_map.end();
       ++it) {

    goto_programt::targett t = it->first;
    num_accelerated += accelerate_loop(t);
  }

  return num_accelerated;
}

bool acceleratet::accelerate(patht &path, goto_programt &accelerator) {
#ifdef DEBUG
  std::cout << "Accelerating:" << endl;

  for (patht::iterator it = path.begin();
       it != path.end();
       ++it) {
    program.output_instruction(ns, "OMG", std::cout, it->loc);
    if(!it->guard.is_nil()) 
      std::cout << "guard: " << from_expr(ns, "OMG", it->guard) << std::endl;
  }
#endif

  polynomial_acceleratort polynomial_accelerator(ns.get_symbol_table(), goto_functions);

  if (polynomial_accelerator.accelerate(path, accelerator)) {
#ifdef DEBUG
    std::cout << "Accelerated!" << endl;
    accelerator.output(ns, "accelerator", std::cout);
#endif

    return true;
  } else {
    return false;
  }
}

void accelerate_functions(
  goto_functionst &functions,
  const namespacet &ns,
  looppath_select_typet looppath_select_type)
{
  Forall_goto_functions (it, functions)
  {
    std::cout << "Accelerating function " << it->first << endl;

    looppath_selectt* looppath_select;
    switch(looppath_select_type)
    {
    case LOOPPATH_SELECT_ALL: 
      looppath_select = new looppath_select_allt(
	 it->second.body, functions, ns,false);
      break;
    case LOOPPATH_SELECT_BLOCKS: 
      looppath_select = new looppath_select_allt(
	 it->second.body, functions, ns,true);
      break;
    case LOOPPATH_SELECT_FILTER: 
      looppath_select = new looppath_select_filtert(
	 it->second.body, functions, ns,false);
      break;
    case LOOPPATH_SELECT_FILTER_BLOCKS: 
      looppath_select = new looppath_select_filtert(
	 it->second.body, functions, ns,true);
      break;
    default: assert(false);
    }
    acceleratet accelerate(it->second.body, functions, ns, *looppath_select);
    int num_accelerated = accelerate.accelerate_loops();

#ifdef DEBUG
    it->second.body.output(ns, "", std::cout);
#endif

    delete looppath_select;

    if (num_accelerated > 0) {
      std::cout << "Added " << num_accelerated << " accelerator(s)" << endl;
    }
  }
}
