#include "dependencies.h"
#include <util/simplify_expr.h>

//get uses
void dependenciest::find_use(goto_programt::targett t,find_symbols_sett& ret) {

  if(t->is_assign()) {
    code_assignt code = to_code_assign(t->code);
    find_symbols(code.rhs(), ret);
  }
  else if(t->is_dead()) {
    code_deadt code = to_code_dead(t->code);
    find_symbols(code.symbol(), ret);
  }
  else if(t->is_goto() || t->is_assume() || t->is_assert()) {
    find_symbols(t->guard, ret);
  }
}

//get defs
void dependenciest::find_def(goto_programt::targett t,find_symbols_sett& ret) {
  if (t->is_assign()) {
    code_assignt code = to_code_assign(t->code);
    find_symbols(code.lhs(), ret);
  }
  /*  else if(t->is_dead()) {
    code_deadt code = to_code_dead(t->code);
    find_symbols(code.symbol(), ret);
    }*/
  else if(t->is_decl()) {
    code_declt code = to_code_decl(t->code);
    find_symbols(code.symbol(), ret);
  }
  else if(t->is_goto() || t->is_assume() || t->is_assert()) {
    find_symbols(t->guard, ret);
  }
}

/******************************************************************************/
//build dependency graph
/******************************************************************************/
void dependenciest::build_graph(goto_programt::targett &start,
				goto_programt::targett &end) 
{
  std::map<goto_programt::targett,find_symbols_sett> target_defs, target_uses;

  branch.clear();
  dep_graph.clear();
  defs.clear();
  uses.clear();

  //compute defs and uses maps
  for(goto_programt::targett it = start; it!=end; it++) {
    find_def(it,target_defs[it]);
    find_use(it,target_uses[it]);
    for(find_symbols_sett::iterator i = target_uses[it].begin(); 
        i!=target_uses[it].end(); i++) {
      uses[*i].insert(it);
    }
    for(find_symbols_sett::iterator i = target_defs[it].begin(); 
          i!=target_defs[it].end(); i++) {
      defs[*i].insert(it);
    }
  }

  for(goto_programt::targett it = start; it!=end; it++) {
    //control flow dependencies
    goto_programt::targett next = it;
    next++;
    if(it->is_goto() && !it->guard.is_constant()) {
      branch[it->get_target()].push_back(brancht(it,true));
      branch[next].push_back(brancht(it,false));
    }

    if(it->is_goto()) {
      branch[it->get_target()].insert(branch[it->get_target()].begin(),
        branch[it].begin(),branch[it].end());
    }
    if(!(it->is_goto() && it->guard.is_constant())) {
      branch[next].insert(branch[next].begin(),branch[it].begin(),branch[it].end());
    }

#if DEBUG
    std::cout << "(" << it->location_number << ") : ";
    output_branches(std::cout,branch[it]);
    std::cout << std::endl; 
#endif

    branchest b; //normalize branch set (to detect joins)
    for(branchest::iterator i1=branch[it].begin();i1!=branch[it].end();i1++)
    {
      bool found = false;
      for(branchest::iterator i2=branch[it].begin();
	    i2!=branch[it].end();i2++)
      {
        if((i1->first == i2->first) && (i1->second != i2->second)) 
        {
	  found = true;  break;          
	}
      }
      for(branchest::iterator i2=b.begin();i2!=b.end();i2++)
      {
        if(*i1==*i2) 
        {
	  found = true;  break;          
	}
      }
      if(!found) b.push_back(*i1);
    }
    branch[it] = b;

#if DEBUG
    std::cout << "(" << it->location_number << ") : ";
    output_branches(std::cout,branch[it]);
    std::cout << std::endl; 
#endif 

    bool independent = true;  
    for(find_symbols_sett::iterator i = target_defs[it].begin(); 
         i!=target_defs[it].end(); i++) {
      if(uses[*i].size()!=0) { independent = false; break; }
    }
    if(independent && (target_uses[it].size()!=0)) independent=false;
    if(independent) continue;   

    for(branchest::iterator i = branch[it].begin(); i!=branch[it].end();i++) {
      dep_graph[it].insert(i->first);
    }

    for(find_symbols_sett::iterator i = target_uses[it].begin(); 
         i!=target_uses[it].end(); i++) {
      //add dependencies to existing defs
      targetst tdef = defs[*i];
      dep_graph[it].insert(tdef.begin(),tdef.end());
    }
    for(find_symbols_sett::iterator i = target_defs[it].begin(); 
         i!=target_defs[it].end(); i++) {
      //add dependencies from existing uses
      targetst tuse = uses[*i];
      for(targetst::iterator t=tuse.begin(); t!=tuse.end(); t++) {
        if(*t != it) {
          dep_graph[*t].insert(it);
	}
      }
    }
  }
}

void dependenciest::output_graph(std::ostream& out) 
{
  for(dep_grapht::iterator t1 = dep_graph.begin(); t1!=dep_graph.end();t1++) 
  {
    out << t1->first->location_number << " --> { ";
    for(targetst::iterator t2 = t1->second.begin(); t2!=t1->second.end();t2++) 
    {
      out << (*t2)->location_number << " ";
    }
    out << "}" << std::endl;
  }
}

void dependenciest::output_cldep_graph(std::ostream& out) 
{
  for(cltarget_mapt::iterator it = cltarget_map.begin();
      it!=cltarget_map.end(); it++) {
    out << it->first << " = { ";
    for(targetst::iterator t = it->second.begin(); t!=it->second.end();t++) 
      out << (*t)->location_number << " ";
    out << "}" << std::endl;
  }
  for(cldep_grapht::iterator it = cldep_graph.begin();
      it!=cldep_graph.end(); it++) {
    out << "(" << it->first << "," << it->second << ")" << std::endl;
  }
}

/******************************************************************************/
//computer cluster graph (Tarjan SCCs)
/******************************************************************************/
void dependenciest::scc_rec(goto_programt::targett n, targetst &visited, 
	     std::list<goto_programt::targett> &stack, 
             unsigned &index,
             std::map<goto_programt::targett, unsigned> &indices,
             std::map<goto_programt::targett, unsigned> &lowlinks)
{
  visited.insert(n);
  stack.push_back(n);
  index++;
  indices[n] = lowlinks[n] = index;   
  for(targetst::iterator it = dep_graph[n].begin();
      it != dep_graph[n].end(); it++) {
    if(visited.find(*it)==visited.end()) {
      scc_rec(*it,visited,stack,index,indices,lowlinks);
      lowlinks[n] = std::min(lowlinks[*it],lowlinks[n]);
    }
    else {
      bool found = false;
      for(std::list<goto_programt::targett>::iterator i = stack.begin(); 
          i!=stack.end(); i++) {
	if(*i==*it) { found = true; break; }
      }
      if(found) {
        lowlinks[n] = std::min(indices[*it],lowlinks[n]);
      }
    }
  }
  if(lowlinks[n]!=indices[n]) return;
  goto_programt::targett t;
  do 
  {
    t = stack.back();
    cltarget_map[lowlinks[n]].insert(t);
    cltarget_maprev[t] = lowlinks[n];
    stack.pop_back();
  }
  while(t!=n);
}

void dependenciest::cluster() {
  cltarget_map.clear();
  cldep_graph.clear();

  targetst targets;
  for(dep_grapht::iterator n = dep_graph.begin();
      n != dep_graph.end(); n++) {
    targets.insert(n->first);
    for(targetst::iterator it = n->second.begin();
	it != n->second.end(); it++) {
      targets.insert(*it);
    }
  }

  targetst visited;
  std::list<goto_programt::targett> stack;
  unsigned index = 0;
  std::map<goto_programt::targett, unsigned> indices;
  std::map<goto_programt::targett, unsigned> lowlinks;
  for(targetst::iterator it = targets.begin(); 
      it!=targets.end(); it++) {
    if(visited.find(*it)!=visited.end()) continue;
    scc_rec(*it,visited,stack,index,indices,lowlinks);
  }
 
  //cltarget_map contains clusters, now reconstruct dependencies
  for(dep_grapht::iterator n = dep_graph.begin();
      n != dep_graph.end(); n++) {
    unsigned cln = cltarget_maprev[n->first];
    for(targetst::iterator it = n->second.begin();
	it != n->second.end(); it++) {
      cldep_graph.insert(cldep_grapht::value_type(cln,cltarget_maprev[*it]));
    }
  }

  //transitive reduction
  reduce();
}

/******************************************************************************/
//transitive reduction of cluster graph
/******************************************************************************/
void dependenciest::reduce() {
  //n->n
  for(cltarget_mapt::iterator n = cltarget_map.begin();
      n != cltarget_map.end(); ++n) {
    cldep_grapht::value_type e(n->first,n->first);
    if(cldep_graph.find(e)!=cldep_graph.end()) cldep_graph.erase(e);
  }

  //n2->n1->n3
  for(cltarget_mapt::iterator n1 = cltarget_map.begin();
      n1 != cltarget_map.end(); ++n1) {
    for(cltarget_mapt::iterator n2 = cltarget_map.begin();
	n2 != cltarget_map.end(); ++n2) {
      for(cltarget_mapt::iterator n3 = cltarget_map.begin(); 
	  n3 != cltarget_map.end(); ++n3) {
	cldep_grapht::value_type e1(n2->first,n3->first);
	cldep_grapht::value_type e2(n2->first,n1->first);
	cldep_grapht::value_type e3(n1->first,n3->first);
	if(cldep_graph.find(e1)!=cldep_graph.end() &&
	   cldep_graph.find(e2)!=cldep_graph.end() &&
	   cldep_graph.find(e3)!=cldep_graph.end()) {
	  cldep_graph.erase(e1);
	}
      }
    }
  }
}


void dependenciest::compress(unsigned min_size) {
  std::set<unsigned> removed;
  for(cltarget_mapt::iterator n = cltarget_map.begin();
      n != cltarget_map.end(); ++n) {
    if(n->second.size()<min_size) {
      //update cltarget_map
      std::set<unsigned> newsuccs;
      cldep_grapht removelinks;
      for(dependenciest::cldep_grapht::iterator it = cldep_graph.begin(); 
	  it != cldep_graph.end(); it++) {  
        if(it->first==(unsigned)-1) continue; 
        if(it->first==n->first) {
          removelinks.insert(*it);
          newsuccs.insert(it->second);
          for(targetst::iterator i = n->second.begin(); i!=n->second.end();i++) {
            cltarget_map[it->second].insert(*i);
            cltarget_maprev[*i] = it->second;
	  }
	}
      }
      if(newsuccs.empty()) continue;      
      removed.insert(n->first);
      //update predecessor links
      cldep_grapht newlinks;
      for(dependenciest::cldep_grapht::iterator it = cldep_graph.begin(); 
	  it != cldep_graph.end(); it++) { 
        if(it->first==(unsigned)-1) continue; 
        if(it->second==n->first) {
          removelinks.insert(*it);
          for(std::set<unsigned>::iterator p = newsuccs.begin();
	      p!=newsuccs.end(); p++) {
            newlinks.insert(std::pair<unsigned,unsigned>(it->first,*p));
	  }
	}
      }
      cldep_graph.insert(newlinks.begin(),newlinks.end());
      //cleanup
      for(cldep_grapht::iterator r = removelinks.begin();
          r!=removelinks.end(); r++) {
	cldep_graph.erase(*r);
      }
    }
  }
  for(std::set<unsigned>::iterator r = removed.begin();
          r!=removed.end(); r++) {
    cltarget_map.erase(*r);
  }

}

dependenciest::branchest dependenciest::negate_branches(const branchest &branches)
{
  branchest res;
  for(branchest::const_iterator it = branches.begin(); it != branches.end(); it++)
  {
    res.push_back(brancht(it->first,!it->second));
  }
  return res;
}

void dependenciest::output_branch(std::ostream& out,const brancht &branch)
{
  out << branch.first->location_number << "," << branch.second << " ";
}

void dependenciest::output_branches(std::ostream& out,const branchest &branches)
{
  for(branchest::const_iterator i = branches.begin(); 
      i!=branches.end(); i++) 
  {
    output_branch(out,*i);
  }
}

dependenciest::branchest dependenciest::diff_branches(const branchest &branches1, 
					    const branchest &branches2)
{
  branchest res;
  for(branchest::const_iterator it1 = branches1.begin(); it1 != branches1.end(); it1++)
  {
    bool found = false;
    for(branchest::const_iterator it2 = branches2.begin(); it2 != branches2.end(); it2++)
    {
      if(*it1==*it2)  { found = true; break; }
    }
    if(!found) res.push_back(*it1);
  }
  return res;
}
