#ifndef CPROVER_ACCELERATE_DEPENDENCIES_H
#define CPROVER_ACCELERATE_DEPENDENCIES_H

#include <util/namespace.h>
#include <iostream>
#include <goto-programs/goto_functions.h>

#include "util.h"

#include <map> 

class dependenciest {
 public:
  dependenciest(goto_programt &_program,
              goto_functionst &_goto_functions,
 	      const namespacet &_ns) :
      program(_program),
      goto_functions(_goto_functions),
      ns(_ns)
  {
  }

  void build_graph(goto_programt::targett &start,
                   goto_programt::targett &end);
  void build_graph() { 
    goto_programt::targett start = program.instructions.begin();
    goto_programt::targett end = program.instructions.end();
    build_graph(start,end); 
  }

  void cluster();
  void compress(unsigned min_size);

  void output_graph(std::ostream& out);
  void output_cldep_graph(std::ostream& out);

  typedef std::pair<goto_programt::targett,bool> brancht;
  typedef std::list<brancht> branchest;

  static branchest negate_branches(const branchest &branches);
  static branchest diff_branches(const branchest &branches1, 
                                      const branchest &branches2);
  static void output_branch(std::ostream& out,const brancht &branch);
  static void output_branches(std::ostream& out,const branchest &branches);

  const branchest& get_branches(goto_programt::targett t) 
    { return branch[t]; }

  unsigned get_cluster_by_target(goto_programt::targett t) 
    { return cltarget_maprev[t]; }

  typedef std::map<goto_programt::targett,targetst> dep_grapht;
  typedef std::set<std::pair<unsigned,unsigned> > cldep_grapht; //cluster deps

  dep_grapht dep_graph;
  cldep_grapht cldep_graph;

 protected:
  goto_programt &program;
  goto_functionst &goto_functions;
  const namespacet &ns;

  void find_def(goto_programt::targett t,find_symbols_sett& ret);
  void find_use(goto_programt::targett t,find_symbols_sett& ret);
  void scc_rec(goto_programt::targett n, targetst &visited, 
	     std::list<goto_programt::targett> &stack, 
             unsigned &index,
	     std::map<goto_programt::targett,unsigned> &indices,
	     std::map<goto_programt::targett,unsigned> &lowlinks);
  void reduce();

  typedef std::map<irep_idt,targetst> var_mapt;
  typedef std::map<unsigned,targetst> cltarget_mapt;
  typedef std::map<goto_programt::targett,unsigned> cltarget_maprevt;

  var_mapt defs,uses;  
  std::map<goto_programt::targett,branchest> branch;
  cltarget_mapt cltarget_map;
  cltarget_maprevt cltarget_maprev;
};


#endif
