#include "loop_parallelizer.h"
#include <goto-programs/goto_program.h>
#include <util/simplify_expr.h>

void loop_parallelizert::do_it(goto_programt::targett &start,
			       goto_programt::targett &end, unsigned min_size) {

  //compute dependency graph
  deps.build_graph(start,end);

#ifdef DEBUG
  std::cout << "dep_graph: " << std::endl;
  deps.output_graph(std::cout);
#endif

  //compute clusters
  deps.cluster();
  deps.compress(min_size);

#ifdef DEBUG
  deps.output_cldep_graph(std::cout);
#endif
  
  //transform body
  result.clear(); 
  transform(start,end);
  
#if 1
  std::cout << "transformed loop body:" << std::endl;
  result.output(ns, "", std::cout);
#endif

}

exprt loop_parallelizert::make_guard(const dependenciest::brancht &b) 
{
  exprt g = b.first->guard;
  if(!b.second) g = not_exprt(g);
  simplify_expr(g,ns);
  return g;
}

exprt loop_parallelizert::make_guard(const dependenciest::branchest &b) 
{
  exprt::operandst c;
  for(dependenciest::branchest::const_iterator it = b.begin(); it != b.end(); it++) 
  {
    c.push_back(make_guard(*it));
  }
  return conjunction(c);
}

void loop_parallelizert::transform(goto_programt::targett &start,
                   goto_programt::targett &end) 
{
  result.instructions.push_back(*start);
  result.instructions.push_back(*end);
  goto_programt::targett afterend = result.add_instruction(SKIP);
  goto_programt::targett curr = result.instructions.begin();
  goto_programt::targett theend = result.instructions.begin(); theend++;
  theend->targets.clear(); theend->targets.push_back(curr);

  std::map<unsigned,goto_programt> blocks;
  typedef std::pair<unsigned,unsigned> edget;
  std::map<edget,goto_programt::targett> dep_map;
  std::map<unsigned,std::list<std::pair<dependenciest::brancht, goto_programt::targett> > >
    branches; //stack of branches and their associated branch targets

  //generate blocks and reconstruct their control flow
  goto_programt::targett it;
  for(it = start; it!=end; it++) 
  {
    unsigned cl = deps.get_cluster_by_target(it);
    if(cl==0) continue;
    if(it->is_goto() && it->get_target()->location_number > end->location_number) //exit loop
    {     
      goto_programt::instructiont i = *it;
      i.targets.clear();  
      i.targets.push_back(afterend); 
      blocks[cl].instructions.push_back(i);

#ifdef DEBUG
      std::cout << "block " << cl << ": " << it->location_number << std::endl;
#endif

    }
    if(!it->is_goto()) 
    {

#ifdef DEBUG
      std::cout << "block " << cl << ": " << it->location_number << std::endl;
#endif

      if(branches.find(cl)==branches.end())  //first branch in block
      {
        blocks[cl].add_instruction(SKIP); //end of block
      }

#ifdef DEBUG
      //print prefixes
      std::cout << "branch prefixes: "; 
      for(std::list<std::pair<dependenciest::brancht, goto_programt::targett> >::iterator 
          br_it = branches[cl].begin(); br_it != branches[cl].end(); br_it++)
      {  
        dependenciest::output_branch(std::cout,br_it->first); 
      }
      std::cout << std::endl;  
#endif

      dependenciest::branchest br0 = deps.get_branches(it);
      //chop off prefix
      dependenciest::branchest br;
      dependenciest::branchest::iterator br_it1 = br0.begin();
      std::list<std::pair<dependenciest::brancht, goto_programt::targett> >::iterator br_it2;
      for(br_it2 = branches[cl].begin(); br_it1 != br0.end(); br_it1++, br_it2++)
      {
        if(*br_it1!=br_it2->first) break;
      }
      for(;br_it1 != br0.end(); br_it1++) br.push_back(*br_it1);

      if(br.empty()) 
      { 
        if(br0.size()==branches[cl].size()) //still the same branch
	{
          goto_programt::targett blbr = branches[cl].back().second;

#ifdef DEBUG
  	  std::cout << "same branch: "; 
          dependenciest::output_branch(std::cout,branches[cl].back().first); 
	  std::cout <<  std::endl; 
#endif

          blocks[cl].instructions.insert(blbr,*it); //insert instruction
	}
        else //some old branch
	{
          goto_programt::targett blbr = (--br_it2)->second;

#ifdef DEBUG
  	  std::cout << "old branch: "; 
          dependenciest::output_branch(std::cout,br_it2->first); 
	  std::cout <<  std::endl; 
#endif

          blocks[cl].instructions.insert(blbr,*it); //insert instruction
	}
      }
      else { //new branch

#ifdef DEBUG
	std::cout << "new branch: "; dependenciest::output_branches(std::cout,br); 
        std::cout << std::endl;
#endif

        //search for branch prefix
        bool neg_of_branch_prefix = false;
        goto_programt::targett prend = blocks[cl].instructions.end(); prend--;
        dependenciest::branchest brneg = dependenciest::negate_branches(br);

        std::list<std::pair<dependenciest::brancht, goto_programt::targett> >::iterator br_it;
        for(br_it = branches[cl].begin(); br_it != branches[cl].end(); br_it++)
	{  
          prend = br_it->second;
          if(br_it->first == brneg.front()) 
	  {

#ifdef DEBUG
  	    std::cout << "matches: "; 
            dependenciest::output_branch(std::cout,br_it->first); 
            std::cout << std::endl;
#endif
            neg_of_branch_prefix = true;
            break;
	  }
	}
        branches[cl].erase(br_it,branches[cl].end()); // pop closed branches

        assert(br.size()!=0);
        if(!neg_of_branch_prefix) //not a negation of a branch prefix
	{
	  //generate branches
	  goto_programt::targett blbr = prend;
          for(dependenciest::branchest::iterator b_it = br.begin();
              b_it != br.end(); b_it++)
	  {
	    blbr = blocks[cl].insert_before(blbr); //end of branch
	    blbr->make_skip();
	    exprt guard = not_exprt(make_guard(*b_it));
	    guard = simplify_expr(guard,ns);
	    if(!guard.is_false()) //jump to end of branch
	      {
		goto_programt::targett goto0 = blocks[cl].insert_before(blbr); 
		goto0->make_goto();
		goto0->targets.push_back(blbr);
		goto0->guard = guard;
	      }
	    //push branch on stack of branches we are in
	    branches[cl].push_back(
	      std::pair<dependenciest::brancht, goto_programt::targett>(*b_it,blbr));
	  }
	  blocks[cl].instructions.insert(blbr,*it); //insert instruction
	}
        else //negation of branch prefix
	{
	  goto_programt::targett  blbr = blocks[cl].insert_after(prend); //end of branch
	  blbr->make_skip();
	  goto_programt::targett goto0 = 
	    blocks[cl].insert_before(prend); //jump from end of then-branch to end
	  goto0->make_goto();
          goto0->targets.push_back(blbr);
          goto0->guard = true_exprt();

          //push first
          dependenciest::branchest::iterator b_it = br.begin();
	    branches[cl].push_back(
	      std::pair<dependenciest::brancht, goto_programt::targett>(*b_it,blbr));
	  //generate branches for rest
          for(b_it++; b_it != br.end(); b_it++)
	  {
	    blbr = blocks[cl].insert_before(blbr); //end of branch
	    blbr->make_skip();
	    exprt guard = not_exprt(make_guard(*b_it));
	    guard = simplify_expr(guard,ns);
	    if(!guard.is_false()) //jump to end of branch
	      {
		goto_programt::targett goto0 = blocks[cl].insert_before(blbr); 
		goto0->make_goto();
		goto0->targets.push_back(blbr);
		goto0->guard = guard;
	      }
	    //push branch on stack of branches we are in
	    branches[cl].push_back(
	      std::pair<dependenciest::brancht, goto_programt::targett>(*b_it,blbr));
	  }
	  blocks[cl].instructions.insert(blbr,*it); //insert instruction
        }
      }
    }
  }
   
  //compute set of successors (independent blocks in the first step)
  std::set<unsigned> cls;
  for(std::map<unsigned,goto_programt>::iterator it = blocks.begin();
      it != blocks.end(); it++) 
  {
    bool independent = true;
    for(dependenciest::cldep_grapht::iterator i = deps.cldep_graph.begin(); 
	i != deps.cldep_graph.end(); i++) 
    {  
      if(i->first==it->first) { independent = false; break; }
    }
    if(independent) cls.insert(it->first); 
  }

  std::set<unsigned> blocks_visited;
  while(cls.size()!=0) {
    //add blocks
    for(std::set<unsigned>::iterator clit = cls.begin();
      clit!=cls.end();clit++) 
    {

#ifdef DEBUG
      std::cout << "block: " << *clit << std::endl;
#endif

      result.destructive_insert(theend,blocks[*clit]);
      blocks_visited.insert(*clit);
      curr++; //points to first instruction in block

      //update predecessor links
      for(dependenciest::cldep_grapht::iterator i = deps.cldep_graph.begin(); 
   	   i != deps.cldep_graph.end(); i++) 
      {  
        if(i->first==*clit) //predecessor
        { 

#ifdef DEBUG
	  std::cout << "update pred (" << i->first << "," << i->second << ") -> " << 
            curr->location_number << std::endl;
#endif

          dep_map[*i]->targets.clear(); 
	  dep_map[*i]->targets.push_back(curr);

#ifdef DEBUG
	  std::cout << dep_map[*i]->location_number << ": goto " << 
            dep_map[*i]->get_target()->location_number << std::endl;
#endif
	}
      }

      //create nondet branchings to succs
      bool first = true; bool none = true;
      for(dependenciest::cldep_grapht::iterator i = deps.cldep_graph.begin(); 
   	   i != deps.cldep_graph.end(); i++) 
      {  
        if(i->second==*clit) { //successor
          none = false;
          if(first) 
          {
	    goto_programt::targett goto0 = result.insert_before(theend);
	    goto0->make_goto();
            goto0->targets.clear();
	    dep_map[*i] = goto0;
	    curr = goto0;
            first = false;
	  }
          else 
          {
	    goto_programt::targett goto1 = result.insert_before(curr);
	    goto1->make_goto();
	    goto1->guard = side_effect_expr_nondett(bool_typet());
            goto1->targets.clear();
	    dep_map[*i] = goto1;
	  }
	}
      }
      if(none)  //final join
      {
	goto_programt::targett goto0 = result.insert_before(theend);
	goto0->make_goto();
        goto0->targets.clear();
        goto0->targets.push_back(theend);
        curr = goto0;

#ifdef DEBUG
        std::cout << curr->location_number << ": goto " << 
          curr->get_target()->location_number << std::endl;
#endif
      }
    }

    //compute directly dependent blocks
    std::set<unsigned> newcls;
    for(dependenciest::cldep_grapht::iterator i = deps.cldep_graph.begin(); 
	i != deps.cldep_graph.end(); i++) {  
      if(cls.find(i->second)!=cls.end()) {
        bool transitive = false;
        for(dependenciest::cldep_grapht::iterator it = deps.cldep_graph.begin();   
            it != deps.cldep_graph.end(); it++) {  
          if(it->first==i->first && 
             blocks_visited.find(i->second)==blocks_visited.end()) {
            transitive = true; break;
	  }
	}
        if(!transitive) newcls.insert(i->first);
      }
    }
    cls = newcls;
  }

  result.update();
}

