#include "looppath_select_filter.h"

//#define DEBUG

//primitive filter for basic integer arithmetic
bool looppath_select_filtert::expr_contains_int_arith_rec(const exprt& expr)
{
  if(expr.id()==ID_plus || expr.id()==ID_minus) return true;
  else 
    forall_operands(it,expr) 
      if(expr_contains_int_arith_rec(*it)) return true;
  return false;
}

bool looppath_select_filtert::filter_int_arith(goto_programt::targett &t) {
  if(t->is_assign()) {
    const code_assignt& c = to_code_assign(t->code); 
    if(c.lhs().type().id()==ID_unsignedbv || 
         c.lhs().type().id()==ID_signedbv)
    {
      if(expr_contains_int_arith_rec(c.rhs())) return true;
    }
  }
  return false;
}

void looppath_select_filtert::find_accelerable_targets(
     goto_programt &program, 
     targetst &targets) {
  Forall_goto_program_instructions(it, program) 
  {
    if(filter_int_arith(it))
    {

#ifdef DEBUG
      std::cout << "TARGET: ";
      program.output_instruction(ns,"",std::cout,it) << std::endl;
#endif

      targets.insert(it);
    }
  }
}

void looppath_select_filtert::find_paths(goto_programt::targett loop_header,
                  natural_loops_mutablet::natural_loopt loop,
                  pathst &loop_paths) {

  //  natural_loops_mutablet::natural_loopt loop;

  targetst targets;

  if(parallelize) {
    goto_programt::targett last_instr_in_loop = loop_header;
    targetst prevh = loop_header->incoming_edges;
    for(targetst::iterator it = prevh.begin(); it!=prevh.end(); it++) {
      if((*it)->is_goto() && (*it)->get_target()==loop_header && 
          loop.find(*it)!=loop.end()) 
	last_instr_in_loop = *it;
    }
    loop_parallelizer.do_it(loop_header,last_instr_in_loop,min_size);
    loop_header = loop_parallelizer.result.instructions.begin();
    loop.clear();
    for(goto_programt::targett it = loop_header;
	it != --loop_parallelizer.result.instructions.end(); it++) {
      loop.insert(it);
    }

    find_accelerable_targets(loop_parallelizer.result,targets);
  }
  else find_accelerable_targets(program,targets);
  //  else loop = _loop;

  std::cout << targets.size() << " accelerable target(s) found " << std::endl;

  patht empty_path;
  while(!targets.empty())
  {
    goto_programt::targett t = *(targets.begin());
    targets.erase(targets.begin());

    pathst prefixes;
    extend_path_backward(t,loop_header,loop,empty_path,prefixes,targets);

#ifdef DEBUG
    std::cout << "backward: " << prefixes.size() << std::endl;
#endif

    pathst new_loop_paths;
    extend_path_forward(++t,loop_header,loop,empty_path,new_loop_paths,targets);

#ifdef DEBUG
    std::cout << "forward: " << new_loop_paths.size() << std::endl;
#endif

    //splice pre- and suffixes
    for (pathst::iterator pit = prefixes.begin(); pit != prefixes.end(); ++pit) 
    {
      if(pit->size()==0) continue;
      for (pathst::iterator it = new_loop_paths.begin(); 
        it != new_loop_paths.end(); ++it) 
      {
        if(it->size()==0) continue;
        patht path = *pit;
        patht suffix = *it;
        path.splice(path.end(),suffix);
	loop_paths.push_back(path);
      }
    }
  }
}

void looppath_select_filtert::extend_path_forward(goto_programt::targett &t,
                              goto_programt::targett &loop_header,
                              natural_loops_mutablet::natural_loopt &loop,
                              patht &prefix,
                              pathst &loop_paths,
                              targetst &targets) 
{
#if 0
  std::cout << "forward: " << t->location_number << std::endl;
#endif

  targetst::iterator tpos = targets.find(t);
  if(tpos!=targets.end()) targets.erase(tpos);

  if (t == loop_header) { // extended prefix might be empty
    // We've expanded a path that has returned to the loop header -- 
    // store this as a looping path.
    patht path(prefix);
    path.push_back(path_nodet(t));
    loop_paths.push_back(path);
    return;
  }

  if (loop.find(t) == loop.end()) {
    // We've dropped out of the loop 
    return;
  }

  // We're still in the loop -- see what we need to do to extend the path.
  if (t->is_goto()) {
    goto_programt::targett next = t;
    ++next;

    patht taken_prefix(prefix);
    patht not_taken_prefix(prefix);
    if(!t->guard.is_constant())
    {
      taken_prefix.push_back(path_nodet(t, t->guard));
      not_taken_prefix.push_back(path_nodet(t, not_exprt(t->guard)));
    }
    //else ; skip unconditional gotos

    for (goto_programt::targetst::iterator it = t->targets.begin();
         it != t->targets.end();
         ++it) {
      goto_programt::targett x = *it;

      if (x->location_number < t->location_number && x != loop_header) {
        // This is a back edge that isn't to the loop header -- it's an inner
        // loop, which we're just ignoring for the moment...
        // TODO: deal with nested loops
        throw "nested loops not supported";
      }

      extend_path_forward(x, loop_header, loop, taken_prefix, 
         loop_paths, targets);
    }

    if(!t->guard.is_constant())
      extend_path_forward(next, loop_header, loop, not_taken_prefix, 
          loop_paths, targets);
  } else if (t->is_assert()) {
    // The assertion passing is a looping path.
    goto_programt::targett next = t;
    ++next;

    patht pass_prefix(prefix);
    pass_prefix.push_back(path_nodet(t, t->guard));

    extend_path_forward(next, loop_header,loop,pass_prefix,loop_paths,targets);

  } else {
    goto_programt::targetst succs;
    program.get_successors(t, succs);

    patht new_prefix(prefix);
    new_prefix.push_back(path_nodet(t));

    for (goto_programt::targetst::iterator it = succs.begin();
         it != succs.end(); ++it) {
      extend_path_forward(*it, loop_header, loop,new_prefix,loop_paths,targets);
    }
  }
}

void looppath_select_filtert::extend_path_backward(goto_programt::targett &t,
                              goto_programt::targett &loop_header,
                              natural_loops_mutablet::natural_loopt &loop,
                              patht &suffix,
                              pathst &paths,
                              targetst &targets) 
{
#if 0
  std::cout << "backward: " << t->location_number << std::endl;
#endif

  targetst::iterator tpos = targets.find(t);
  if(tpos!=targets.end()) targets.erase(tpos);

  if (t == loop_header && !suffix.empty()) {
    // We've expanded a path that has reached the loop header
    patht path(suffix);
    path.push_front(path_nodet(t));
    paths.push_back(path);
    return;
  }

  if (loop.find(t) == loop.end()) {
    std::cout << "target " << t->location_number << std::endl;
    // We've dropped out of the loop -- this should not happen
    assert(false);
  }

  // We're still in the loop -- see what we need to do to extend the path.
  patht new_suffix(suffix);
  new_suffix.push_front(path_nodet(t));
  for (std::set<goto_programt::targett>::iterator it = t->incoming_edges.begin();
         it != t->incoming_edges.end();
         ++it) 
  {
    goto_programt::targett prev_inst = *it;
    extend_path_backward(prev_inst, loop_header, loop, new_suffix, paths, targets);
  }
}
