/*******************************************************************\

Module: Branch Instrumentation

Author: Daniel Kroening, kroening@kroening.com

\*******************************************************************/

#include <util/cprover_prefix.h>
#include <util/std_code.h>
#include <util/std_expr.h>
#include <util/i2string.h>
#include <util/simplify_expr.h>
#include <set>
#include <iostream>
#include <string>

#include "branchtest.h"

/*******************************************************************\

Function: branch

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void branchtest(
  symbol_tablet &symbol_table,
  goto_functionst &goto_functions)
{
  namespacet ns(symbol_table);
  unsigned assertion_counter = 0;
  Forall_goto_functions(f_it, goto_functions)
  {
    // don't instrument our internal functions
    if(has_prefix(id2string(f_it->first), CPROVER_PREFIX))
      continue;
    
    // patch in a call to `id' at the branch points
    goto_programt &body=f_it->second.body;
    
    Forall_goto_program_instructions(i_it, body)
    {
      // if C goto T is transformed into:
      //
      // if !C goto T'          i_it
      // assert(true);           t1
      // goto T                 t2
      // T': assert(true);   t3
      // ...
    
      if(!i_it->is_goto()) continue;
      exprt guard = simplify_expr(i_it->guard,ns);
      if(!guard.is_constant())
      {
        // negate condition
        i_it->guard.make_not();
        
        goto_programt::targett t1=body.insert_after(i_it);
        t1->make_assertion(true_exprt()); //put "true" assertion as marker
        t1->location.set_comment(
          (TESTGOAL_PREFIX "::branch")+i2string(assertion_counter++));
        
        goto_programt::targett t2=body.insert_after(t1);
        t2->make_goto();
        t2->targets=i_it->targets;
        
        goto_programt::targett t3=body.insert_after(t2);
        t3->make_assertion(true_exprt()); //put "true" assertion as marker
        t3->location.set_comment(
          (TESTGOAL_PREFIX "::branch")+i2string(assertion_counter++));
        i_it->targets.clear();
        i_it->targets.push_back(t3);
      }    
    }
  }
}
